# -*- coding: utf-8 -*-

from textwrap import dedent

from webargs import fields as wfields
from flask_apispec import (marshal_with, use_kwargs as use_args,
                           MethodResource as Resource)
from flask_apispec.annotations import doc
import numpy as np
import pandas as pd
from sklearn.metrics import (adjusted_rand_score,
                             adjusted_mutual_info_score,
                             precision_score, recall_score, f1_score,
                             v_measure_score)
from sklearn.metrics import (roc_auc_score, average_precision_score)
import warnings
from sklearn.exceptions import UndefinedMetricWarning

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.ingestion import _check_mutual_index
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.engine.categorization import _CategorizerWrapper
from freediscovery.utils import _docstring_description, _paginate
from freediscovery.cluster import centroid_similarity, compute_optimal_sampling
from freediscovery.engine.cluster import _ClusteringWrapper
from freediscovery.engine.search import _SearchWrapper
from freediscovery.metrics import (categorization_score,
                       ratio_duplicates_score, f1_same_duplicates_score,
                       mean_duplicates_count_score, _scale_cosine_similarity)
from freediscovery.engine.near_duplicates import _DuplicateDetectionWrapper
from freediscovery.engine.email_threading import _EmailThreadingWrapper
from freediscovery.datasets import load_dataset
from freediscovery.exceptions import WrongParameter
from freediscovery.engine.stop_words import _StopWordsWrapper
from .validators import _is_in_range

from .schemas import (IDSchema, FeaturesParsSchema,
                      FeaturesSchema, _DatasetDefinitionShort,
                      DocumentIndexNestedSchema, _DatasetDefinition,
                      ExampleDatasetSchema, DocumentIndexFullSchema,
                      LsiParsSchema, LsiPostSchema,
                      _CategorizationInputSchema,
                      CategorizationParsSchema, CategorizationPostSchema,
                      CategorizationPredictSchema, ClusteringSchema,
                      _CategorizationIndex, ErrorSchema,
                      _CategorizationPredictSchemaElement,
                      MetricsCategorizationSchema, MetricsClusteringSchema,
                      MetricsDupDetectionSchema,
                      EmailThreadingSchema, EmailThreadingParsSchema,
                      SearchResponseSchema,
                      EmptySchema,
                      CustomStopWordsSchema, CustomStopWordsLoadSchema
                      )

EPSILON = 1e-3  # small numeric value

# =========================================================================== #
#                        Datasets download                                    #
# =========================================================================== #


class ExampleDatasetApi(Resource):

    @use_args({'n_categories': wfields.Int(missing=2)})
    @doc(description=_docstring_description(dedent(load_dataset.__doc__)))
    @marshal_with(ExampleDatasetSchema())
    def get(self, name, **args):
        n_categories = args['n_categories']
        n_categories = min(3, n_categories)
        n_categories = max(1, n_categories)

        categories = None
        if "20_newsgroups_" in name:
            if n_categories == 3:
                categories = ['comp.graphics', 'rec.sport.baseball',
                              'sci.space']
            elif n_categories == 2:
                categories = ['comp.graphics', 'rec.sport.baseball']
            elif n_categories == 1:
                categories = ['comp.graphics']

        md, training_set, test_set = load_dataset(name, self._cache_dir,
                                                  verbose=True,
                                                  verify_checksum=False,
                                                  categories=categories)

        return {'metadata': md, 'training_set': training_set,
                'dataset': test_set}


# Definine the response formatting schemas
id_schema = IDSchema()
features_schema = FeaturesSchema()
error_schema = ErrorSchema()

# ========================================================================== #
#                    Feature extraction                                      #
# ========================================================================== #


class FeaturesApi(Resource):

    @doc(description='View parameters used for the feature extraction')
    @marshal_with(FeaturesSchema(many=True))
    def get(self):
        fe = FeatureVectorizer(self._cache_dir)
        return fe.list_datasets()

    @doc(description=dedent("""
            Initialize the feature extraction on a document collection.

            **Parameters**
             - `n_features`: [optional] number of features (overlapping character/word n-grams that are hashed).  n_features refers to the number of buckets in the hash.  The larger the number, the fewer collisions.   (default: 1100000)
             - `analyzer`: 'word', 'char', 'char_wb' Whether the feature should be made of word or character n-grams.  Option ‘char_wb’ creates character n-grams only from text inside word boundaries.  ( default: 'word')
             - `ngram_range` : tuple (min_n, max_n), default=(1, 1) The lower and upper boundary of the range of n-values for different n-grams to be extracted. All values of n such that min_n <= n <= max_n will be used.

             - `stop_words`: "english" or "None" Remove stop words from the resulting tokens. Only applies for the "word" analyzer.  If "english", a built-in stop word list for English is used. ( default: "english")
             - `n_jobs`: The maximum number of concurrently running jobs (default: 1)
             - `chuck_size`: The number of documents simultaneously processed by a running job (default: 5000)
             - `weighting`: the SMART notation for document term weighting and normalization.  In the form [nlabL][ntp][ncb] , see https://en.wikipedia.org/wiki/SMART_Information_Retrieval_System
             - `norm_alpha`: the alpha value used for pivoted normalization

             - `use_hashing`: Enable hashing. This option must be set to True for classification and set to False for clustering. (default: True)
             - `min_df`: When building the vocabulary ignore terms that have a document frequency strictly lower than the given threshold. This value is ignored when hashing is used.
             - `max_df`: When building the vocabulary ignore terms that have a document frequency strictly higher than the given threshold. This value is ignored when hashing is used.
             - `parse_email_headers`: when documents are emails, attempt to parse the information contained in the header (default: False)
             - `preprocess`: a list of pre-processing steps to apply before vectorization. A subset of ['emails_ignore_header'], default: [].
             - `id`: (optional) custom dataset id. Can only contain letters, numbers, "_" or "-". It must also be between 2 and 50 characters long.
             - `overwrite`: if a custom dataset id was provided, and it already exists, overwrite it. Default: false
            """))
    @use_args(FeaturesParsSchema(strict=True, exclude=('data_dir')))
    @marshal_with(IDSchema())
    def post(self, **args):
        if args['use_hashing']:
            for key in ['min_df', 'max_df']:
                if key in args:
                    # the above parameters are ignored with caching
                    del args[key]
        for key in ['min_df', 'max_df']:
            if key in args and args[key] > 1. + EPSILON:  # + eps
                args[key] = int(args[key])
        dsid = args.pop('id', None)
        overwrite = args.pop('overwrite', None)

        if dsid is None:
            mode = 'w'
        elif not overwrite:
            mode = 'w'
        else:
            mode = 'fw'

        fe = FeatureVectorizer(self._cache_dir, mode=mode, dsid=dsid)
        dsid = fe.setup(**args)
        return {'id': dsid}


class FeaturesApiElement(Resource):
    @doc(description="Load extracted features "
                     "(and obtain the processing status)")
    def get(self, dsid):
        sc = FeaturesSchema()
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid)
        out = fe.pars_.copy()
        is_processing = (fe.cache_dir / dsid / 'processing').exists()
        is_finished = (fe.cache_dir / dsid / 'processing_finished').exists()
        if is_processing and not is_finished:
            n_chunks = len((fe.cache_dir / dsid).glob('features-*[0-9]'))
            out['n_samples_processed'] = min(n_chunks*out['chunk_size'],
                                             out['n_samples'])
            return sc.dump(out).data, 202
        elif not is_processing and is_finished:
            out['n_samples_processed'] = out['n_samples']
            out['filenames'] = fe.filenames_
            return sc.dump(out).data, 200
        else:
            return error_schema.dump({"message":
                                     "Processing failed, see server logs!"
                                      }).data, 520

    @doc(description=dedent("""
         Run feature extraction on a dataset,

         **Parameters**
          - `data_dir`: [optional] relative path to the directory with the input files. Either `data_dir` or `dataset_definition` must be provided.
          - `dataset_definition`: [optional] a list of dictionaries `[{'file_path': <str>, 'content': <str>, 'document_id': <int>, 'rendition_id': <int>}, ...]` where `document_id` and `rendition_id` are optional, while either `file_path` or `content` field must be provided. 
          - `vectorize`: [optional] this option can be used to ingest the dataset_definition in batches (optionally with document content), then make one final call to vectorize all sent documents (bool, default: True)
          - `document_id_generator`: [optional] if the `document_id` is not provided, this specifies how it is generated. If `indexed_file_path` the `document_id` is given by the index of the sorted `file_path`, otherwise if `infer_file_path` the `document_id` is inferred from the `file_path` strings, removing all non digit characters. In this second case, the `file_path` must contain a unique numeric ID (default: `indexed_file_path`)
         """))
    @use_args({"data_dir":  wfields.Str(),
               "dataset_definition": wfields.Nested(_DatasetDefinition, many=True),
               "vectorize": wfields.Bool(missing=True),
               "document_id_generator": wfields.Str(missing="indexed_file_path")})
    @marshal_with(IDSchema())
    def post(self, dsid, **args):
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid, mode='r')
        fe.ingest(**args)
        if fe.pars_['parse_email_headers']:
            fe.parse_email_headers()
        return {'id': fe.dsid}

    @doc(description='Delete a processed dataset')
    @marshal_with(EmptySchema())
    def delete(self, dsid):
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid)
        fe.delete()
        return {}


class FeaturesApiElementMappingNested(Resource):
    @doc(description=dedent("""
         Compute correspondence between id fields for documents.
         At least one of the fields used for indexing must be provided,
         and all the rest will be computed (if available).
         If the data parameter is not provided, return all the correspondence table

         **Parameters**
          - `data`: the ids of documents used as the query
          - `return_file_path`: whether the results should include the file path
         """))
    @use_args({'data': wfields.Nested(DocumentIndexFullSchema, many=True),
               'return_file_path': wfields.Bool(missing=True)})
    @marshal_with(DocumentIndexNestedSchema())
    def post(self, dsid, **args):
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid)
        if 'data' in args and args['data']:
            query = pd.DataFrame(args['data'])
            fe.db_.filenames_ = fe.filenames_
            res = fe.db_.search(query, return_file_path=args['return_file_path'])
        else:
            res = None
        if args['return_file_path']:
            fe.db_.filenames_ = fe.filenames_
        res_repr = fe.db_.render_dict(res, return_file_path=args['return_file_path'])
        return {'data': res_repr}


class FeaturesApiAppend(Resource):
    @doc(description=dedent("""
         Add new documents to an existing processed dataset.
         This will also automatically update the LSI model if any
         is present. Raw documents on disk are not affected.

         This operation cannot be undone.

         Warning: all categorization, clustering, duplicate detection and
         email threading models associated with this dataset will be removed and
         need to be re-trained.

         **Parameters**
          - `data_dir`: [optional] relative path to the directory with the input files. Either `data_dir` or `dataset_definition` must be provided.
          - `dataset_definition`: [optional] a list of dictionaries `[{'file_path': <str>, 'document_id': <int>, 'rendition_id': <int>}, ...]` where  `rendition_id` are optional, while either `file_path` or `content` field must be provided. 
          """))
    @use_args({'data_dir': wfields.Str(),
               'dataset_definition': wfields.Nested(_DatasetDefinition, many=True,
                                                    required=True)})
    @marshal_with(EmptySchema())
    def post(self, dsid, **args):
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid)
        fe.append(**args)
        return {}


class FeaturesApiRemove(Resource):
    @doc(description=dedent("""Remove documents from an existing processed dataset.
         This will also automatically update the LSI model if any
         is present.  Raw documents on disk are not affected.

         This operation cannot be undone.

         Warning: all categorization, clustering, duplicate detection and
         email threading models associated with this dataset will be removed and
         need to be re-trained.

         **Parameters**
          - `dataset_definition`: [optional] a list of dictionaries `[{'file_path': <str>, 'document_id': <int>, 'rendition_id': <int>}, ...]` where  `rendition_id` are optional.
          """))
    @use_args({'dataset_definition': wfields.Nested(_DatasetDefinitionShort,
                                                    many=True, required=True)})
    @marshal_with(EmptySchema())
    def post(self, dsid, **args):
        fe = FeatureVectorizer(self._cache_dir, dsid=dsid, mode='r')
        fe.remove(args['dataset_definition'])
        return {}


# =========================================================================== #
#                 LSI decomposition
# =========================================================================== #


class LsiApi(Resource):

    @doc(description='List existing LSI models')
    @use_args({'parent_id': wfields.Str(required=True)})
    @marshal_with(LsiParsSchema(many=True))
    def get(self, **args):
        parent_id = args['parent_id']
        lsi = _LSIWrapper(cache_dir=self._cache_dir, parent_id=parent_id,
                          random_state=self._random_seed)
        return lsi.list_models()

    @doc(description=dedent("""
           Build a Latent Semantic Indexing (LSI) model

           Recommended data ingestion options also include, `use_idf=1, sublinear_tf=0, binary=0`.

           The recommended value for the `n_components` (dimensions of the SVD decompositions) is
           in the [100, 200] range.

           **Parameters**
             - `n_components`: Desired dimensionality of the output data. Must be strictly less than the number of features.
             - `parent_id`: parent dataset identified by `dataset_id`
             - `alpha`: floor on the number of components used with small datasets
             - `id`: (optional) custom model id. Can only contain letters, numbers, "_" or "-". It must also be between 2 and 50 characters long.
             - `overwrite`: if a custom model id was provided, and it already exists, overwrite it. Default: false

          """))
    @use_args({'parent_id': wfields.Str(required=True),
               'n_components': wfields.Int(missing=150),
               'alpha': wfields.Number(missing=0.33),
               'id': wfields.Str(),
               'overwrite': wfields.Boolean(missing=False)})
    @marshal_with(LsiPostSchema())
    def post(self, **args):
        parent_id = args.pop('parent_id')

        mid = args.pop('id', None)
        overwrite = args.pop('overwrite', None)
        if mid is None:
            mode = 'w'
        elif not overwrite:
            mode = 'w'
        else:
            mode = 'fw'

        lsi = _LSIWrapper(cache_dir=self._cache_dir, parent_id=parent_id,
                          random_state=self._random_seed, mid=mid, mode=mode)
        _, explained_variance = lsi.fit_transform(**args)
        return {'id': lsi.mid, 'explained_variance': explained_variance}


class LsiApiElement(Resource):

    @doc(description='Show Latent Semantic Indexing (LSI) model parameters')
    @marshal_with(LsiParsSchema())
    def get(self, mid):
        cat = _LSIWrapper(self._cache_dir, mid=mid)

        pars = cat._load_pars()
        pars['parent_id'] = pars['parent_id']
        return pars

    @doc(description='Delete a Latent Semantic Indexing (LSI) model')
    @marshal_with(EmptySchema())
    def delete(self, mid):
        cat = _LSIWrapper(self._cache_dir, mid=mid)
        cat.delete()
        return {}

# =========================================================================== #
#                 Categorization (ML)
# =========================================================================== #


class ModelsApi(Resource):
    @doc(description='List existing categorization models')
    @marshal_with(CategorizationParsSchema(many=True))
    def get(self, parent_id):
        cat = _CategorizerWrapper(parent_id, self._cache_dir)

        return cat.list_models()

    @doc(description=dedent("""
           Build the categorization ML model

           The option `use_hashing=True` must be set for the feature extraction. Recommended options also include, `weighting="ntc"`.

           **Parameters**
            - `parent_id`: `dataset_id` or `lsi_id`
            - `data`: a list of dict which have a `category` field and one or several fields that can be used for indexing, such as `document_id` and optionally `rendition_id`.
            - `method`: classification algorithm to use (default: LogisticRegression),
              * "LogisticRegression": [LogisticRegression](http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html#sklearn.linear_model.LogisticRegression)
              * "LinearSVC": [Linear SVM](http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html),
              * "NearestNeighbor": nearest neighbor classifier (requires LSI)
            - `cv`: binary, if true optimal parameters of the ML model are determined by cross-validation over 5 stratified K-folds (default False).
            - `training_scores`: binary, compute the efficiency scores on the training dataset. This would make computations much slower for NearestNeighbors (default False). 
          """))
    @use_args(_CategorizationInputSchema())
    @marshal_with(CategorizationPostSchema())
    def post(self, **args):
        training_scores = args['training_scores']
        parent_id = args['parent_id']
        cat = _CategorizerWrapper(self._cache_dir, parent_id=parent_id,
                                  random_state=self._random_seed)

        query = pd.DataFrame(args['data'])
        res_q = cat.fe.db_.search(query, drop=False)
        del args['data']

        args['index'] = res_q.internal_id.values
        args['y'] = res_q.category.values

        if args['cv']:
            cv = 'fast'
        else:
            cv = None
        for key in ['parent_id', 'cv', 'training_scores']:
            del args[key]
        _, Y_train = cat.fit(cv=cv, **args)
        idx_train = args['index']
        res = {'id': cat.mid, 'training_scores': {}}
        if training_scores:
            Y_res, md = cat.predict()
            idx_res = np.arange(cat.fe.n_samples_, dtype='int')
            res['training_scores'] = categorization_score(idx_train, Y_train,
                                                          idx_res, np.argmax(Y_res, axis=1))
        return res


class ModelsApiElement(Resource):
    @doc(description='Load categorization model parameters')
    @marshal_with(CategorizationParsSchema())
    def get(self, mid):
        cat = _CategorizerWrapper(self._cache_dir, mid=mid)
        pars = cat.get_params()
        return pars

    @doc(description='Delete the categorization model')
    @marshal_with(EmptySchema())
    def delete(self, mid):
        cat = _CategorizerWrapper(self._cache_dir, mid=mid)
        cat.delete()
        return {}


class ModelsApiPredict(Resource):

    @doc(description=dedent("""
            Predict document categorization with a previously trained model

            **Parameters**
             - `max_result_categories` : the maximum number of categories in the results
             - `sort_by` : if provided and not None, the field used for sorting results. Valid values are [None, 'score'] or any of the ingested category names.
             - `sort_order`: the sort order (if applicable), one of ['ascending', 'descending']
             - `max_results` : return only the first `max_results` documents. If `max_results <= 0` all documents are returned.
             - `ml_output` : type of the output in ['decision_function', 'probability'], only affects ML methods.
             - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
             - `min_score` : filter out results below a similarity threashold
             - `subset`: apply prediction to a document subset. Must be one of ['all', 'train', 'test']. Default: 'test'.
             - `subset_document_id`: apply prediction to a subset of document_id. 
             - `batch_id`: retrieve a given subset of scores (-1 to retrieve all). Default: 0
             - `batch_size`: the number of document scores retrieved per batch. Default: 10000
            """))
    @use_args({'max_result_categories': wfields.Int(missing=1),
               'sort_by': wfields.Str(missing='score'),
               'sort_order': wfields.Str(missing='descending',
                                         validate=_is_in_range(['descending',
                                                                'ascending'])),
               'max_results': wfields.Int(),
               'ml_output': wfields.Str(missing='probability'),
               'metric': wfields.Str(missing='cosine'),
               'min_score': wfields.Number(missing=-1),
               'subset': wfields.Str(missing='test',
                                     validate=_is_in_range(['all', 'train',
                                                            'test'])),
               'subset_document_id': wfields.List(wfields.Int()),
               'batch_id': wfields.Int(missing=0),
               'batch_size': wfields.Int(missing=10000),
               })
    @marshal_with(CategorizationPredictSchema())
    def get(self, mid, **args):

        valid_sort = ['score']

        sort_by = args.pop('sort_by')
        sort_ascending = args.pop('sort_order') == 'ascending'
        max_result_categories = args.pop('max_result_categories')
        min_score = args.pop("min_score")
        max_results = args.pop("max_results", 0)
        subset = args.pop("subset")
        subset_document_id = args.pop('subset_document_id', None)
        batch_size = args.pop("batch_size")
        batch_id = args.pop("batch_id")

        cat = _CategorizerWrapper(self._cache_dir, mid=mid)
        y_res, nn_res = cat.predict(**args)
        train_indices = cat._pars['index']

        labels = list(cat.le.classes_)
        Y_pred = y_res

        if max_result_categories <= 0:
            raise WrongParameter(('the max_result_categories={} '
                                  'must be strictly positive')
                                 .format(max_result_categories))
        if sort_by not in valid_sort + labels:
            raise WrongParameter(("sort_by={} not value. Must be "
                                  "one of {}")
                                 .format(sort_by, valid_sort + labels))

        id_mapping = cat.fe.db_.data
        # have to cast to object as otherwise
        # we get serializing np.int64 issues...
        base_keys = [key for key in id_mapping.columns
                     if key in ['internal_id', 'document_id', 'rendition_id']]
        id_mapping = id_mapping[base_keys].set_index('internal_id', drop=True).astype('object')
        # create dataframe out of results
        Y_pred = id_mapping.copy()
        for idx, el in enumerate(labels):
            Y_pred[el] = y_res[:, idx]
        if nn_res is not None:
            nn_range = np.arange(nn_res.shape[0])
            NN_map = []
            for idx, el in enumerate(labels):
                NN_map_el = pd.DataFrame({'internal_id': nn_res[:, idx]},
                                         index=nn_range)
                NN_map_el = NN_map_el.join(id_mapping, on='internal_id',
                                           how='left')
                NN_map.append(NN_map_el)
        else:
            NN_map = None

        # optionally filter out test or training set
        if subset_document_id is not None:
            Y_pred = Y_pred.reset_index().set_index('document_id', drop=False)
            subset_document_id = np.array(subset_document_id)
            subset_mask = np.in1d(Y_pred.index, subset_document_id)
            Y_pred = Y_pred.iloc[subset_mask].set_index('internal_id')
        elif subset in ['train', 'test']:
            _mask = np.in1d(Y_pred.index.values, train_indices)
            if subset == 'test':
                _mask = ~_mask
            Y_pred = Y_pred.loc[_mask, :]

        # sort output
        if sort_by in labels:
            Y_pred = Y_pred.sort_values(sort_by, ascending=sort_ascending)

        Y_pred_max = Y_pred[labels].max(axis=1)
        if sort_by in valid_sort:
            if sort_by == 'score':
                Y_pred_max = Y_pred_max.sort_values(ascending=sort_ascending)

        # filter out low scores
        if min_score is not None:
            Y_pred_max = Y_pred_max[Y_pred_max > min_score]

        Y_pred = Y_pred.loc[Y_pred_max.index.values, :]

        # return only first N results
        if max_results > 0:
            Y_pred = Y_pred.iloc[:max_results]

        Y_pred, pagination = _paginate(Y_pred, batch_id, batch_size)

        # render NN mapping
        if NN_map is not None:
            for idx, NN_map_el in enumerate(NN_map):
                NN_map[idx] = NN_map_el.loc[Y_pred.index, :].to_dict(orient='records')

        res = Y_pred.to_dict(orient='records')

        # convert scores to a more usable format
        for idx, row in enumerate(res):
            scores = []
            for label_id, key in enumerate(labels):
                scores_el = {'category': key,
                             'score': row.pop(key)}
                if NN_map is not None:
                    scores_el.update(NN_map[label_id][idx])
                scores.append(scores_el)
            row['scores'] = scores

        def sort_func(scores):
            return -scores['score']

        # sort categories and return only the first max_result_categories
        for idx, row in enumerate(res):
            row['scores'] = sorted(row['scores'], key=sort_func)[:max_result_categories]

        return {'data': res, 'pagination': pagination}

# =========================================================================== #
#                             Clustering
# =========================================================================== #


class KmeanClusteringApi(Resource):

    @doc(description=dedent("""
           Compute K-mean clustering

           The option `use_hashing=False` must be set for the feature extraction. Recommended options for feature extraction include, `weighting="ntc"`.

           **Parameters**
            - `parent_id`: `dataset_id` or `lsi_id`
            - `n_clusters`: the number of clusters
            - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
           """))
    @use_args({'parent_id': wfields.Str(required=True),
               'n_clusters': wfields.Int(missing=150),
               'metric': wfields.Str(missing='cosine')})
    @marshal_with(IDSchema())
    def post(self, **args):
        metric = args.pop('metric')

        cl = _ClusteringWrapper(cache_dir=self._cache_dir,
                                parent_id=args['parent_id'],
                                metric=metric)

        del args['parent_id']

        cl.k_means(**args)
        return {'id': cl.mid}


class BirchClusteringApi(Resource):

    @doc(description=dedent("""
           Compute birch clustering

           The option `use_hashing=False` must be set for the feature extraction. Recommended options for data ingestion also include, `ntc`.

           **Parameters**
            - `parent_id`: `dataset_id` or `lsi_id`
            - `n_clusters`: the number of clusters or -1 to use hierarchical clustering (default: -1)
            - `min_similarity`: The radius of the subcluster obtained by merging a new sample and the closest subcluster should be lesser than the threshold. Otherwise a new subcluster is started. See [sklearn.cluster.Birch](http://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html). Increasing this value would increase the hierarchical tree depth (and the number of clusters).
            - `branching_factor`: Maximum number of CF subclusters in each node. If a new samples enters such that the number of subclusters exceed the branching_factor then the node has to be split. The corresponding parent also has to be split and if the number of subclusters in the parent is greater than the branching factor, then it has to be split recursively. Decreasing this value would increase the number of clusters.
            - `max_tree_depth` : Maximum hierarchy depth (only applicable when `n_clusters=-1`)
            - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
           """))
    @use_args({
            'parent_id': wfields.Str(required=True),
            'n_clusters': wfields.Int(missing=-1),
            'branching_factor': wfields.Int(missing=20),
            'max_tree_depth': wfields.Int(),
            # this corresponds approximately to threashold = 0.5
            'min_similarity': wfields.Number(missing=0.5),
            'metric': wfields.Str(missing='cosine')
            }
            )
    @marshal_with(IDSchema())
    def post(self, **args):
        from math import sqrt

        metric = args.pop('metric')
        S_cos = _scale_cosine_similarity(args.pop('min_similarity'),
                                         metric=metric,
                                         inverse=True)
        # cosine sim to euclidean distance
        threshold = sqrt(2 * (1 - S_cos))

        cl = _ClusteringWrapper(cache_dir=self._cache_dir,
                                parent_id=args.pop('parent_id'),
                                metric=metric)

        if args.get('n_clusters') <= 0:
            args['n_clusters'] = None

        cl.birch(threshold=threshold, **args)
        return {'id': cl.mid}


class DBSCANClusteringApi(Resource):

    @doc(description=dedent("""
           Compute clustering (DBSCAN)

           The option `use_hashing=False` must be set for the feature extraction. Recommended options for the data ingestion also include, `weighting="ntc"`.

           **Parameters**
             - `parent_id`: `dataset_id` or `lsi_id`
             - `min_similarity`: The radius of the subcluster obtained by merging a new sample and the closest subcluster should be lesser than the threshold. Otherwise a new subcluster is started. See [sklearn.cluster.Birch](http://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html)
             - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
             - `min_samples`: (optional) int The number of samples (or total weight) in a neighborhood for a point to be considered as a core point. This includes the point itself.
            """))
    @use_args({'parent_id': wfields.Str(required=True),
               'min_samples': wfields.Int(missing=10),
               # this corresponds approximately to threashold = 0.5
               'min_similarity': wfields.Number(missing=0.5),
               'metric': wfields.Str(missing='cosine')})
    @marshal_with(IDSchema())
    def post(self, **args):
        from math import sqrt
        metric = args.pop('metric')
        S_cos = _scale_cosine_similarity(args.pop('min_similarity'),
                                         metric=metric,
                                         inverse=True)
        # cosine sim to euclidean distance
        eps = sqrt(2 * (1 - S_cos))

        cl = _ClusteringWrapper(cache_dir=self._cache_dir,
                                parent_id=args.pop('parent_id'),
                                metric=metric)

        cl.dbscan(eps=eps, **args)
        return {'id': cl.mid}


class ClusteringApiElement(Resource):

    @doc(description=dedent("""
           Compute cluster labels

           **Parameters**
            - `n_top_words`: keep only most relevant `n_top_words` words
            - `return_optimal_sampling` : Instead of cluster results, the optimal sampling results will be returned (with no cluster labels). This option is only valid with Birch algorithm. Note that optimal sampling cannot return more samples than the subclusters in the birch clustering results (default: false)
            - `sampling_min_similarity` : Similarity threashold used by smart sampling. Decreasing this value would result in more sampled documents. Default: 1.0 (i.e. use the full cluster hierarichy).
            - `sampling_min_coverage` : Minimal coverage requirement in [0, 1] range. Increasing this value would result in a larger number of samples. (default: 0.9)
            """))
    @use_args({'n_top_words': wfields.Int(missing=5),
               'return_optimal_sampling': wfields.Bool(missing=False),
               'sampling_min_similarity': wfields.Number(missing=1.0),
               'sampling_min_coverage': wfields.Number(missing=0.9)})
    @marshal_with(ClusteringSchema())
    def get(self, method, mid, **args):
        return_optimal_sampling = args.pop('return_optimal_sampling')
        sampling_min_coverage = args.pop('sampling_min_coverage')
        sampling_min_similarity = args.pop('sampling_min_similarity')

        cl = _ClusteringWrapper(cache_dir=self._cache_dir, mid=mid)
        km = cl._load_model()

        if return_optimal_sampling and not cl._pars['is_hierarchical']:
            raise WrongParameter(('Model {} does not support optimal sampling,'
                                  'please use hierarchical Birch clustering '
                                  '(with n_clusters=-1)')
                                 .format(type(km).__name__))

        cl._fit_X = cl.pipeline.data

        if type(km).__name__ in ['_BirchDummy', 'Birch'] and cl._pars['is_hierarchical']:
            # Hierarchical clustering

            htree = cl._load_htree()

            db = cl.fe.db_.data

            if return_optimal_sampling:
                # cut the hierarchical tree to match the smart sampling
                flat_tree = compute_optimal_sampling(htree,
                                                     sampling_min_similarity,
                                                     sampling_min_coverage)
            else:
                # we don't use optimal sampling
                if cl._pars['max_tree_depth'] is not None:
                    max_tree_depth = cl._pars['max_tree_depth']
                    htree.limit_depth(max_tree_depth)
                flat_tree = htree.flatten()

                terms = cl.compute_labels(
                            cluster_indices=[row['document_id_accumulated']
                                             for row in flat_tree],
                            **args)
                for label, row in zip(terms, flat_tree):
                    row['cluster_label'] = label

            res = []
            doc_keys = [key for key in db.columns
                        if key in ['document_id', 'rendering_id']]
            db = db[doc_keys]

            for row in flat_tree:
                irow = {'cluster_similarity': row['cluster_similarity'],
                        'cluster_size': row['cluster_size'],
                        'cluster_id': row['cluster_id']}
                if not return_optimal_sampling:
                    irow['cluster_label'] = ' '.join(row['cluster_label'])
                    irow['children'] = [el['cluster_id']
                                        for el in row.children]
                    irow['cluster_depth'] = row.current_depth

                db_sl = db.iloc[row['document_id_accumulated']].copy()
                db_sl['similarity'] = row['document_similarity']
                tmp = []
                for index, row_tmp in db_sl.iterrows():
                    row_dict = row_tmp.to_dict()
                    tmp.append(row_dict)
                irow['documents'] = tmp

                res.append(irow)

        else:
            # Non hierarchical clustering

            if args['n_top_words'] > 0:
                terms = cl.compute_labels(**args)
            else:
                terms = None

            y = cl._merge_response(km.labels_)
            res = []
            valid_keys = ['document_id', 'rendering_id', 'similarity']
            for name, group in y.groupby('cluster_id'):
                name = int(name)

                S_sim_mean, S_sim = centroid_similarity(cl._fit_X,
                                                        group.index.values,
                                                        cl._pars['metric'])
                group = group.assign(similarity=S_sim)

                row_docs = []
                for idx, row in group.iterrows():
                    row_docs.append({key: val
                                     for key, val in row.to_dict().items()
                                     if key in valid_keys})
                row['documents'] = row_docs
                irow = {'documents': row_docs, 'cluster_id': int(name),
                        'cluster_similarity': S_sim_mean}
                irow['cluster_size'] = len(row_docs)
                if terms is not None:
                    irow['cluster_label'] = ' '.join(terms[name])
                res.append(irow)

        return {'data': res}

    @doc(description='Delete a clustering model')
    @marshal_with(EmptySchema())
    def delete(self, method, mid):  # TODO unused parameter 'method'
        cl = _ClusteringWrapper(cache_dir=self._cache_dir, mid=mid)
        cl.delete()
        return {}

# =========================================================================== #
#                             Duplicate detection
# =========================================================================== #


class DupDetectionApi(Resource):

    @doc(description=dedent("""
           Compute near duplicates

           **Parameters**
            - `parent_id`: `dataset_id` or `lsi_id`
            - `method`: str, default='simhash' Method used for duplicate detection. One of "simhash", "i-match"
          """))
    @use_args({"parent_id": wfields.Str(required=True),
               "method": wfields.Str(required=False, missing='simhash')})
    @marshal_with(IDSchema())
    def post(self, **args):

        model = _DuplicateDetectionWrapper(cache_dir=self._cache_dir,
                                           parent_id=args['parent_id'])

        del args['parent_id']

        model.fit(args['method'])

        return {'id': model.mid}


class DupDetectionApiElement(Resource):

    @doc(description=dedent("""
           Query duplicates

           **Parameters**
            - `distance` : int, default=2 Maximum number of differnet bits in the simhash (Simhash method only)
            - `n_rand_lexicons` : int, default=1 number of random lexicons used for duplicate detection (I-Match method only)
            - `rand_lexicon_ratio` : float, default=0.7 ratio of the vocabulary used in random lexicons (I-Match method only)
            - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
          """))
    @use_args({'distance': wfields.Int(),
               'n_rand_lexicons': wfields.Int(),
               'rand_lexicon_ratio': wfields.Number(),
               'metric': wfields.Str(missing='cosine')
               })
    @marshal_with(ClusteringSchema())
    def get(self, mid, **args):

        metric = args.pop('metric')

        model = _DuplicateDetectionWrapper(cache_dir=self._cache_dir, mid=mid)
        cluster_id = model.query(**args)
        model._fit_X = model.pipeline.data  # load the data
        y = model._merge_response(cluster_id)
        res = []
        valid_keys = ['document_id', 'rendering_id', 'similarity']
        for name, group in y.groupby('cluster_id'):
            if group.shape[0] <= 1:
                continue

            S_sim_mean, S_sim = centroid_similarity(model._fit_X,
                                                    group.index.values,
                                                    metric)
            group = group.assign(similarity=S_sim)

            row_docs = []
            for idx, row in group.iterrows():
                row_docs.append({key: val for key, val in row.to_dict().items()
                                 if key in valid_keys})
            row['documents'] = row_docs
            res.append({'documents': row_docs, 'cluster_id': name,
                        'cluster_similarity': S_sim_mean})

        return {'data': res}

    @marshal_with(EmptySchema())
    def delete(self, mid):

        model = _DuplicateDetectionWrapper(cache_dir=self._cache_dir, mid=mid)
        model.delete()
        return {}

# =========================================================================== #
#                            Metrics                                          #
# =========================================================================== #


class MetricsCategorizationApiElement(Resource):
    @doc(description=dedent("""
          Compute categorization metrics to assess the quality
          of categorization.

          In the case of binary categrorization, category labels are sorted alphabetically
          and the second one is expected to be the positive one.

          **Parameters**
            - y_true: [required] ground truth categorization data
            - y_pred: [required] predicted categorization results
            - metrics: [required] list of str. Metrics to compute, any combination of "precision", "recall", "f1", "roc_auc"
          """))
    @use_args({'y_true': wfields.Nested(_CategorizationIndex,
                                        many=True, required=True),
               'y_pred': wfields.Nested(_CategorizationPredictSchemaElement,
                                        many=True, required=True),
               'metrics': wfields.List(wfields.Str())})
    @marshal_with(MetricsCategorizationSchema())
    def post(self, **args):
        from sklearn.preprocessing import LabelEncoder
        from ..metrics import recall_at_k_score
        output_metrics = {}
        y_true = pd.DataFrame(args['y_true'])

        y_pred_b = []
        for row in args['y_pred']:
            nrow = {'document_id': row['document_id'],
                    'category': row['scores'][0]['category'],
                    'score': row['scores'][0]['score']}
            y_pred_b.append(nrow)

        y_pred_b = pd.DataFrame(y_pred_b)
        index_cols = _check_mutual_index(y_true.columns, y_pred_b.columns)

        y_true = y_true.set_index(index_cols, verify_integrity=True)
        y_pred_b = y_pred_b.set_index(index_cols, verify_integrity=True)

        le = LabelEncoder()
        # this also sorts label by arithmetic order
        y_true['category_id'] = le.fit_transform(y_true.category.values)
        y_pred_b['category_id'] = le.transform(y_pred_b.category.values)

        y = y_true[['category_id']].merge(y_pred_b[['category_id', 'score']],
                                          how='inner',
                                          left_index=True,
                                          right_index=True,
                                          suffixes=('_true', '_pred'))
        if 'metrics' in args:
            metrics = args['metrics']
        else:
            metrics = ['precision', 'recall', 'roc_auc',
                       'f1', 'average_precision', 'recall_at_k']

        _binary_metrics = ['precision', 'recall', 'f1']

        cy_true = y.category_id_true.values
        cy_pred = y.category_id_pred.values

        n_classes = len(le.classes_)

        if n_classes == 2:
            cy_pred_score = y.score.values
            cy_pred_score[cy_pred == 0] *= -1

        # wrapping metrics calculations, as for example F1 score
        # can frequently print warnings "F-score is ill defined
        # and being set to 0.0 due to no predicted samples"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UndefinedMetricWarning)
            for func in [precision_score,
                         recall_score,
                         f1_score,
                         roc_auc_score,
                         average_precision_score,
                         recall_at_k_score]:
                name = func.__name__.replace('_score', '')
                opts = {}
                if name in ['roc_auc', 'average_precision', 'recall_at_k'] \
                        and n_classes == 2:
                    y_targ = cy_pred_score
                    if name == 'recall_at_k':
                        opts = {'k': 0.2}
                else:
                    y_targ = cy_pred

                if name in _binary_metrics and n_classes != 2:
                    opts['average'] = 'micro'
                if name in metrics:
                    if n_classes == 2 or name in _binary_metrics:
                        output_metrics[name] = func(cy_true, y_targ, **opts)
                    else:
                        output_metrics[name] = np.nan
            if "recall_at_k" in output_metrics:
                output_metrics['recall_at_20p'] = output_metrics.pop('recall_at_k')

        return output_metrics


_metrics_clustering_api_get_args = {
    'labels_true': wfields.List(wfields.Int(), required=True),
    'labels_pred': wfields.List(wfields.Int(), required=True),
    'metrics': wfields.List(wfields.Str(), required=True)
}


class MetricsClusteringApiElement(Resource):
    @doc(description=dedent("""
          Compute clustering metrics to assess the quality
          of categorization, comparing the groud truth labels with the predicted ones.

          **Parameters**
            - labels_true: [required] list of int. Ground truth clustering labels
            - labels_pred: [required] list of int. Predicted clustering labels
            - metrics: [required] list of str. Metrics to compute, any combination of "adjusted_rand", "adjusted_mutual_info", "v_measure"
          """))
    @use_args(_metrics_clustering_api_get_args)
    @marshal_with(MetricsClusteringSchema())
    def post(self, **args):
        output_metrics = dict()
        labels_true = args['labels_true']
        labels_pred = args['labels_pred']
        metrics = args['metrics']

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UndefinedMetricWarning)
            if 'adjusted_rand' in metrics:
                output_metrics['adjusted_rand'] = adjusted_rand_score(labels_true, labels_pred)
            if 'adjusted_mutual_info' in metrics:
                output_metrics['adjusted_mutual_info'] = adjusted_mutual_info_score(labels_true, labels_pred)
            if 'v_measure' in metrics:
                output_metrics['v_measure'] = v_measure_score(labels_true, labels_pred)
        return output_metrics


class MetricsDupDetectionApiElement(Resource):
    @doc(description=dedent("""
          Compute duplicate detection metrics to assess the quality
          of categorization, comparing the groud truth labels with the predicted ones.

          **Parameters**
            - labels_true: [required] list of int. Ground truth clustering labels
            - labels_pred: [required] list of int. Predicted clustering labels
            - metrics: [required] list of str. Metrics to compute, any combination of "ratio_duplicates", "f1_same_duplicates", "mean_duplicates_count"
          """))
    # Arguments are the same as for clustering
    @use_args(_metrics_clustering_api_get_args)
    @marshal_with(MetricsDupDetectionSchema())
    def post(self, **args):
        output_metrics = dict()
        labels_true = args['labels_true']
        labels_pred = args['labels_pred']
        metrics = args['metrics']

        # Methods 'ratio_duplicates_score' and 'f1_same_duplicates_score'
        # in ..metrics.py
        # accept Numpy array objects, not standard Python lists
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UndefinedMetricWarning)
            if 'ratio_duplicates' in metrics:
                output_metrics['ratio_duplicates'] = \
                   ratio_duplicates_score(np.array(labels_true),
                                          np.array(labels_pred))
            if 'f1_same_duplicates' in metrics:
                output_metrics['f1_same_duplicates'] = \
                   f1_same_duplicates_score(np.array(labels_true),
                                            np.array(labels_pred))
            if 'mean_duplicates_count' in metrics:
                output_metrics['mean_duplicates_count'] = \
                    mean_duplicates_count_score(labels_true, labels_pred)
        return output_metrics

# =========================================================================== #
#                             Email threading
# =========================================================================== #


class EmailThreadingApi(Resource):

    @doc(description='Compute email threading')
    @use_args({"parent_id": wfields.Str(required=True)})
    @marshal_with(EmailThreadingSchema())
    def post(self, **args):

        model = _EmailThreadingWrapper(cache_dir=self._cache_dir,
                                       parent_id=args['parent_id'])

        tree = model.thread()

        return {'data': [el.to_dict(include=['subject']) for el in tree],
                'id': model.mid}


class EmailThreadingApiElement(Resource):

    @doc(description='Get email threading parameters')
    @marshal_with(EmailThreadingParsSchema())
    def get(self, mid):

        model = _EmailThreadingWrapper(cache_dir=self._cache_dir, mid=mid)

        return model.get_params()

    @doc(description='Delete a processed dataset')
    @marshal_with(EmptySchema())
    def delete(self, mid):

        model = _EmailThreadingWrapper(cache_dir=self._cache_dir, mid=mid)
        model.delete()
        return {}


# =========================================================================== #
#                             (Semantic) search
# =========================================================================== #


class SearchApi(Resource):
    @doc(description=dedent("""
            Perform document search (if `parent_id` is a `dataset_id`) or a semantic search (if `parent_id` is a `lsi_id`).

            Parameters
            ----------
            - `parent_id` : the id of the previous processing step (either `dataset_id` or `lsi_id`)
            - `query` : the seach query. Either `query` or `query_document_id` must be provided.
            - `query_document_id` : the id of the document used as the search query. Either `query` or `query_document_id` must be provided.
            - `metric` : The similarity returned by nearest neighbor classifier in ['cosine', 'jaccard', 'cosine-positive'].
            - `min_score` : filter out results below a similarity threashold
            - `max_results` : return only the first `max_results` documents. If `max_results <= 0` all documents are returned.
            - `sort_by` : if provided and not None, the field used for sorting results. Valid values are [None, 'score']
            - `sort_order`: the sort order (if applicable), one of ['ascending', 'descending']
            - `batch_id`: retrieve a given subset of scores (-1 to retrieve all). Default: 0
            - `batch_size`: the number of document scores retrieved per batch. Default: 10000
             - `subset_document_id`: apply prediction to a subset of document_id. 
            """))
    @use_args({"parent_id": wfields.Str(required=True),
               "query": wfields.Str(),
               "query_document_id": wfields.Int(),
               'metric': wfields.Str(missing='cosine'),
               'min_score': wfields.Number(missing=-1),
               'max_results': wfields.Int(),
               'sort_by': wfields.Str(missing='score'),
               'sort_order': wfields.Str(missing='descending',
                                         validate=_is_in_range(['descending',
                                                                'ascending'])),
               'batch_id': wfields.Int(missing=0),
               'batch_size': wfields.Int(missing=10000),
               'subset_document_id': wfields.List(wfields.Int()),
               })
    @marshal_with(SearchResponseSchema())
    def post(self, **args):
        parent_id = args['parent_id']
        subset_document_id = args.pop('subset_document_id', None)
        model = _SearchWrapper(cache_dir=self._cache_dir, parent_id=parent_id)

        if 'query' in args and 'query_document_id' not in args:
            query = args['query']
            scores = model.search(query, metric=args['metric'])
        elif 'query' not in args and 'query_document_id' in args:
            query = pd.DataFrame([{'document_id': args['query_document_id']}])
            res_q = model.fe.db_.search(query, drop=False)

            scores = model.search(None,
                                  internal_id=res_q.internal_id.values[0],
                                  metric=args['metric'])
        else:
            raise WrongParameter("One of the 'query', "
                                 "'query_document_id' must be provided")

        scores_pd = pd.DataFrame({'score': scores,
                                  'internal_id': np.arange(
                                       model.fe.n_samples_, dtype='int')})
        scores_pd = scores_pd[scores_pd.score > args['min_score']]

        if 'query' not in args and 'query_document_id' in args:
            # remove the query document
            scores_pd = scores_pd[scores_pd.internal_id != res_q.internal_id.values[0]]

        if subset_document_id is not None:
            db = model.fe.db_.data[['document_id', 'internal_id']].set_index('internal_id')
            scores_pd = scores_pd.set_index('internal_id', drop=False)
            scores_pd = scores_pd.join(db, how='left')
            scores_pd = scores_pd.set_index('document_id', drop=False)
            subset_document_id = np.array(subset_document_id)
            subset_mask = np.in1d(scores_pd.index, subset_document_id)
            scores_pd = scores_pd.iloc[subset_mask].set_index('internal_id', drop=False)

        sort_by = args['sort_by']
        if sort_by:
            if sort_by not in scores_pd.columns:
                raise WrongParameter(
                    'sort_by={} not in {}'.format(sort_by,
                                                  list(scores_pd.columns)))
            is_ascending = args['sort_order'] == 'ascending'
            scores_pd.sort_values(by=sort_by, inplace=True,
                                  ascending=is_ascending)

        if 'max_results' in args and args['max_results'] > 0:
            scores_pd = scores_pd.iloc[:args['max_results'], :]

        # make the pagination happen
        scores_batch, pagination = _paginate(scores_pd, args['batch_id'],
                                             args['batch_size'])

        res = model.fe.db_.render_dict(scores_batch)

        return {'data': res, 'pagination': pagination}


# =========================================================================== #
#                            Custom Stop Words
# =========================================================================== #


class CustomStopWordsApi(Resource):
    @doc(description="Store a list of custom stop words")
    @use_args({"name": wfields.Str(required=True),
               "stop_words": wfields.List(wfields.Str(), required=True)})
    @marshal_with(CustomStopWordsSchema())
    def post(self, **args):
        name = args['name']
        stop_words = args['stop_words']
        model = _StopWordsWrapper(cache_dir=self._cache_dir)
        model.save(name=name, stop_words=stop_words)
        return {'name': name}


class CustomStopWordsLoadApi(Resource):

    @doc(description="Load a stored list of stop words")
    @marshal_with(CustomStopWordsLoadSchema())
    def get(self, name):
        return {'name': name,
                'stop_words': _StopWordsWrapper(
                    cache_dir=self._cache_dir).load(name)}

    @doc(description='Delete a stored custom stop words')
    @marshal_with(EmptySchema())
    def delete(self, name):
        _StopWordsWrapper(cache_dir=self._cache_dir).delete(name)
        return {}



# -*- coding: utf-8 -*-

import os
import sys

from freediscovery.externals.joblib.pool import _get_temp_dir

base_dir = os.path.dirname(os.path.dirname(__file__))

EXTERNAL_DATASETS_PATH = os.environ.get('FREEDISCOVERY_EXTERNAL_TEST_DATASETS')


def run(coverage=False):
    import pytest
    argv = ['-x', base_dir, '-v', '-s']  #, "
    if coverage:
        argv += ["--cov=freediscovery"]
    result = pytest.main(argv)
    status = int(result)
    return status


def run_cli(coverage=False):
    status = run(coverage=coverage)
    print('Exit status: {}'.format(status))
    sys.exit(status)


def check_cache(test_env=True):
    import tempfile

    if test_env:
        subfolder = 'freediscovery-cache-test-{}'.format(os.getpid())
    else:
        subfolder = 'freediscovery-cache'

    cache_dir, _ = _get_temp_dir(subfolder, temp_folder=tempfile.gettempdir())

    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    return cache_dir

if __name__ == '__main__':
    run_cli()



"""
Semantic Search
===============

An example of Semantic Search
"""

from __future__ import print_function

import os.path
import requests
import pandas as pd

pd.options.display.float_format = '{:,.3f}'.format
pd.options.display.expand_frame_repr = False

dataset_name = "treclegal09_2k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

###############################################################################
#
# 0. Load the test dataset
# ------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()

# create a custom dataset definition for ingestion
data_dir = input_ds['metadata']['data_dir']
dataset_definition = [{'document_id': row['document_id'],
                       'file_path': os.path.join(data_dir, row['file_path'])}
                      for row in input_ds['dataset']]

###############################################################################
#
# 1. Feature extraction
# ---------------------
# 1.a Load dataset and initalize feature extraction

url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))

###############################################################################
#
# 1.b Start feature extraction

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
requests.post(url, json={'dataset_definition': dataset_definition})

###############################################################################
#
# 2. Calculate LSI
# ----------------
# (used for Nearest Neighbors method)

url = BASE_URL + '/lsi/'
print("POST", url)
n_components = 100
res = requests.post(url,
                    json={'n_components': n_components,
                          'parent_id': dsid
                          }).json()

lsi_id = res['id']
print('  => LSI model id = {}'.format(lsi_id))
print(("  => SVD decomposition with {} dimensions explaining "
       "{:.2f} % variabilty     of the data")
      .format(n_components, res['explained_variance']*100))


###############################################################################
#
# 3. Semantic search
# ------------------

print("\n3.a. Perform the semantic search")


query = ("There are some conflicts with the draft date, so we will probably "
         "need to have it on a different date.")

url = BASE_URL + '/search/'
print(" POST", url)

res = requests.post(url,
                    json={'parent_id': lsi_id,
                          'query': query
                          }).json()

data = res['data']

df = pd.DataFrame(data).set_index('document_id')
print(df)

print(df.score.max())


###############################################################################
#
# Delete the extracted features

url = BASE_URL + '/feature-extraction/{}'.format(dsid)
requests.delete(url)



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import numpy as np

from sklearn.base import BaseEstimator
from sklearn.neighbors import NearestNeighbors, NearestCentroid
from sklearn.utils.validation import check_array
from freediscovery.base import RankerMixin

# a subclass of the NearestCentroid from scikit-learn that also
# includes the distance to the nearest centroid


class NearestCentroidRanker(NearestCentroid):

    def decision_function(self, X):
        """Compute the distances to the nearest centroid for
        an array of test vectors X.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        Returns
        -------
        C : array, shape = [n_samples]
        """
        from sklearn.metrics.pairwise import pairwise_distances
        from sklearn.utils.validation import check_array, check_is_fitted

        check_is_fitted(self, 'centroids_')

        X = check_array(X, accept_sparse='csr')

        return pairwise_distances(X, self.centroids_,
                                  metric=self.metric).min(axis=1)


def _chunk_kneighbors(func, X, batch_size=5000, **args):
    """ Chunk kneighbors computations to reduce RAM requirements

    Parameters
    ----------
    func : function
      the function to run
    X : ndarray
      the array func is applied to
    batch_size : int
      batch size

    Returns
    -------
    dist : array
       distance array
    ind : array of indices
    """
    n_samples = X.shape[0]
    ind_arr = []
    dist_arr = []
    # don't enter the last loop if n_sampes is a multiple of batch_size
    for k in range(n_samples//batch_size + int(n_samples % batch_size != 0)):
        mslice = slice(k*batch_size, min((k+1)*batch_size, n_samples))
        X_sl = X[mslice, :]
        dist_k, ind_k = func(X_sl, **args)
        ind_arr.append(ind_k)
        dist_arr.append(dist_k)
    return (np.concatenate(dist_arr, axis=0), np.concatenate(ind_arr, axis=0))


class NearestNeighborRanker(BaseEstimator, RankerMixin):
    """A nearest neighbor ranker.

    The distance is returned in terms of cosine_similarity

    Parameters
    ----------
    radius : float, optional (default = 1.0)
        Range of parameter space to use by default for :meth:`radius_neighbors`
        queries.

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDtree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.

    n_jobs : int, optional (default = 1)
        The number of parallel jobs to run for neighbors search.
        If ``-1``, then the number of jobs is set to the number of CPU cores.

    method : str, def
        If "unsupervised" only distances to the positive samples are used
        in the ranking. If "supervised" both the distance to the positive
        and negative documents are used for ranking (i.e. if a document
        is slightly further away from a positive document than from a negative
        one, it will be considered negative with a very low score)

    """

    def __init__(self, radius=1.0, algorithm='brute', leaf_size=30, n_jobs=1):

        self.algorithm = algorithm
        self.radus = radius
        self.n_jobs = n_jobs
        self.leaf_size = leaf_size

    def fit(self, X, y):
        """Fit the model using X as training data
        Parameters
        ----------
        X : {array-like, sparse matrix, BallTree, KDTree}
            Training data, shape [n_samples, n_features],

        """
        X = check_array(X, accept_sparse='csr')
        y = np.asarray(y, dtype='int')
        y_unique = np.unique(y)

        index = np.arange(len(y), dtype='int')

        if len(y_unique) == 0:
            raise ValueError('The training set must have at least '
                             'one document category!')

        # define nearest neighbors search objects for each category
        self._mod = [NearestNeighbors(n_neighbors=1,
                                      leaf_size=self.leaf_size,
                                      algorithm=self.algorithm,
                                      n_jobs=self.n_jobs,
                                      # euclidean metric by default
                                      metric='cosine',
                                      ) for el in range(len(y_unique))]

        index_mapping = []
        for imod, y_val in enumerate(y_unique):
            mask = (y == y_val)
            index_mapping.append(index[mask])
            self._mod[imod].fit(X[mask])

        self.index_mapping = index_mapping

    def kneighbors(self, X=None, batch_size=5000):
        """Finds the K-neighbors of a point.
        Returns indices of and distances to the neighbors of each point.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            the input array
        batch_size : int
            the batch size
        Returns
        -------
        S_cos : array [n_samples, n_categories]
            the cosine similarity to closest point in each category
        indices : array [n_samples, n_categories]
            Indices of the nearest points in the population matrix.
        --------
        """
        X = check_array(X, accept_sparse='csr')

        n_classes = len(self._mod)

        S_res = np.zeros((X.shape[0], n_classes), dtype='float')
        nn_idx_res = np.zeros((X.shape[0], n_classes), dtype='int')

        for imod in range(n_classes):
            D_i, nn_idx_i_loc = _chunk_kneighbors(self._mod[imod].kneighbors,
                                                  X,
                                                  batch_size=batch_size)

            # only NearestNeighbor-1 (only one column in the kneighbors output)
            # convert from eucledian distance in L2 norm space to cosine
            # similarity
            # S_cos = seuclidean_dist2cosine_sim(D_i[:,0])
            S_res[:, imod] = 1 - D_i[:, 0]
            # map local index within index_mapping to global index
            nn_idx_res[:, imod] = self.index_mapping[imod][nn_idx_i_loc[:, 0]]

        return S_res, nn_idx_res



# -*- coding: utf-8 -*-

from __future__ import print_function

from time import time

import numpy as np
import scipy.linalg
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize


n_samples = 20000
n_lsi = 150

np.random.seed(999999)

X = np.random.randn(n_samples, n_lsi)
normalize(X, norm='l2', copy=False)

print('Test set: {} docs, {} LSI dim'.format(n_samples, n_lsi))

for N_train in [1000, 2000]:
    mod = NearestNeighbors(n_neighbors=1, # always at 1 in FreeDiscovery
                          algorithm='ball_tree',
                          radius=1.0,
                          leaf_size=30,
                          n_jobs=1 # value of 1 best for the benchmark
                          )

    t0 = time()
    mod.fit(X[:N_train, :])
    t1 = time()
    mod.kneighbors(X, return_distance=True)
    t2 = time()

    print('Training set {} docs, t_train= {:.2f} s, t_test = {:.2f}s'.format(
                        N_train, t1-t0, t2-t1))




# -*- coding: utf-8 -*-
import platform
import re

import pandas as pd


def parse_ground_truth_file(filename):
    """ Parse a ground truth file specified by a filename.
    Replace '/' by '\' when running in Windows """
    df = pd.read_csv(filename, sep='[\s\t]+',
                     names=['file_path', 'is_relevant'], engine='python')
    if platform.system() == 'Windows':
        df.file_path = df.file_path.map(lambda path: path.replace('/', '\\'))
    return df


def parse_smart_tokens(text):
    """
    Parse a dataset stored in the SMART tokenized format, used
    in particular for the RCV1-v2 dataset,
    http://www.jmlr.org/papers/volume5/lewis04a/lyrl2004_rcv1v2_README.htm
    (cf. Appendix B.12.i.)

    Parameters
    ----------
    text : str
       the full text of the dataset

    Returns
    -------
    result : dict
       the parsed dataset in a OrderedDict, with document_ids as keys,
       and a string of tokens as values
    """

    data = None
    data_all = []
    data_key = None

    for line in text.splitlines():
        key_match = re.match('^\.(?P<key>[A-Z])\s?(?P<val>.*)', line)
        if key_match:
            data_key = key_match.group('key')
            if data_key == 'I':
                if data is not None:
                    data_all.append(data)
                data = {"I": int(key_match.group('val'))}
            else:
                data[data_key] = []
        else:
            if data_key is None:
                raise ValueError('Failed to parse index at the first line!')
            if data_key != 'I':
                data[data_key].append(line)

    if data is not None:
        data_all.append(data)

    for row in data_all:
        for key, val in row.items():
            if key != 'I':
                row[key] = ' '.join(val)

    df = pd.DataFrame(data_all)
    return df.set_index('I')



# -*- coding: utf-8 -*-

import os

import numpy as np
from numpy.testing import assert_allclose

from freediscovery.io import parse_ground_truth_file
from freediscovery.metrics import categorization_score
from freediscovery.tests.run_suite import check_cache


basename = os.path.dirname(__file__)

cache_dir = check_cache()

data_dir = os.path.join(basename, "..", "data", "ds_001", "raw")

ground_truth = parse_ground_truth_file(
                  os.path.join(data_dir, "..", "ground_truth_file.txt"))


def test_unique_label():
    """Check that testing works with only one label in the training test"""
    np.random.seed(10)
    Nshape = ground_truth.file_path.values.shape
    is_relevant = np.zeros(Nshape).astype(int)

    idx = np.arange(len(is_relevant), dtype='int')

    categorization_score(idx, is_relevant, idx, np.random.rand(*Nshape))


def test_categorization_score():
    idx = [1, 2,  3,  4,  5, 6]
    y = [1, 1, -1, -1, -1, 1]
    idx_ref = [10, 5, 3, 2, 6]
    y_ref = [0,  1, 0, 1, 1]

    scores = categorization_score(idx_ref, y_ref, idx, y)

    assert_allclose(scores['precision'], 1.0)
    assert_allclose(scores['recall'], 0.66666666, rtol=1e-4)

    # make sure permutations don't affect the result
    idx_ref2 = [10, 5, 2, 3, 6]
    y_ref2 = [0, 1, 1, 0, 1]
    scores2 = categorization_score(idx_ref2, y_ref2, idx, y)
    assert scores['average_precision'] == scores2['average_precision']



# -*- coding: utf-8 -*-

import os

import numpy as np
from unittest import SkipTest
from numpy.testing import assert_allclose, assert_equal, assert_array_equal
import pytest

from sklearn.preprocessing import normalize
from sklearn.exceptions import NotFittedError
from sklearn.datasets import make_blobs

from freediscovery.cluster import select_top_words
from freediscovery.cluster.hierarchy import _check_birch_tree_consistency
from freediscovery.cluster import compute_optimal_sampling, centroid_similarity
from freediscovery.cluster import Birch, birch_hierarchy_wrapper



NCLUSTERS = 2

@pytest.mark.parametrize('dataset, optimal_sampling',
                         [('random', False),
                          ('birch_hierarchical', False),
                          ('birch_hierarchical', True)])
def test_birch_make_hierarchy(dataset, optimal_sampling):

    if dataset == 'random':
        np.random.seed(9999)

        X = np.random.rand(1000, 100)
        normalize(X)
        branching_factor = 10
    elif dataset == 'birch_hierarchical':
        basename = os.path.dirname(__file__)
        X = np.load(os.path.join(basename, '..', 'data',
                    'ds_lsi_birch', 'data.npy'))
        branching_factor = 2

    mod = Birch(n_clusters=None, threshold=0.1,
                branching_factor=branching_factor, compute_labels=False,
                compute_sample_indices=True)
    mod.fit(X)

    htree, n_subclusters = birch_hierarchy_wrapper(mod)

    # let's compute cluster similarity
    for row in htree.flatten():
        inertia, S_sim = centroid_similarity(X,
                                             row['document_id_accumulated'])
        row['document_similarity'] = S_sim
        row['cluster_similarity'] = inertia

    assert htree.tree_size == n_subclusters

    doc_count = 0
    for el in htree.flatten():
        doc_count += len(el['document_id'])
        el.current_depth
        el.document_id_accumulated
    assert doc_count == len(htree['document_id_accumulated'])
    assert doc_count == X.shape[0]
    assert htree.document_count == X.shape[0]
    if optimal_sampling:
        s_samples_1 = compute_optimal_sampling(htree, min_similarity=0.85,
                                               min_coverage=0.9)

        for row in s_samples_1:
            assert len(row['document_similarity']) == 1
            assert len(row['document_id_accumulated']) == 1
        s_samples_2 = compute_optimal_sampling(htree, min_similarity=0.85,
                                               min_coverage=0.2)
        s_samples_3 = compute_optimal_sampling(htree, min_similarity=0.9,
                                               min_coverage=0.9)

        assert len(s_samples_1) > len(s_samples_2)
        assert len(s_samples_1) < len(s_samples_3)


def test_birch_hierarchy_fitted():
    model = Birch()

    with pytest.raises(NotFittedError):
        birch_hierarchy_wrapper(model)


def test_birch_hierarchy_validation():
    with pytest.raises(ValueError):
        birch_hierarchy_wrapper("some other object")


@pytest.mark.parametrize('example_id', [12, 34])
def test_birch_example_reproducibility(example_id):
    # check reproducibility of the Birch example
    rng = np.random.RandomState(42)

    X, y = make_blobs(n_samples=1000, n_features=10, random_state=rng)

    cluster_model = Birch(threshold=0.9, branching_factor=20,
                          compute_sample_indices=True)
    cluster_model.fit(X)
    # assert len(cluster_model.root_.subclusters_[1].child_.subclusters_) == 3

    htree, n_subclusters = birch_hierarchy_wrapper(cluster_model)

    assert htree.tree_size == n_subclusters

    # same random seed as in the birch hierarchy example
    assert htree.tree_size == 78
    sc = htree.flatten()[example_id]
    if example_id == 34:
        # this is true in both cases, but example_id fails on circle ci
        assert sc.current_depth == 1
        assert len(sc.children) == 3

    assert_array_equal([sc['cluster_id'] for sc in htree.flatten()],
                       np.arange(htree.tree_size))


def test_denrogram_children():
    # temporary solution for
    # https://stackoverflow.com/questions/40239956/node-indexing-in-hierarachical-clustering-dendrograms
    import numpy as np
    from scipy.cluster.hierarchy import dendrogram, linkage
    from freediscovery.cluster import _DendrogramChildren

    # generate two clusters: a with 10 points, b with 5:
    np.random.seed(1)
    a = np.random.multivariate_normal([10, 0], [[3, 1], [1, 4]],
                                      size=[10, ])
    b = np.random.multivariate_normal([0, 20], [[3, 1], [1, 4]],
                                      size=[5, ])
    X = np.concatenate((a, b),)
    Z = linkage(X, 'ward')
    # make distances between pairs of children uniform
    # (re-scales the horizontal (distance) axis when plotting)
    Z[:, 2] = np.arange(Z.shape[0])+1

    ddata = dendrogram(Z, no_plot=True)
    dc = _DendrogramChildren(ddata)
    idx = 0
    # check that we can compute children for all nodes
    for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
        node_children = dc.query(idx)
        idx += 1
    # last level node should encompass all samples
    assert len(node_children) == X.shape[0]
    assert_allclose(sorted(node_children), np.arange(X.shape[0]))


def test_dbscan_noisy_utils():
    from freediscovery.cluster.utils import (_dbscan_noisy2unique,
                                             _dbscan_unique2noisy)
    from sklearn.metrics import v_measure_score

    x_ref = np.array([-1, 0, -1,  1, 1, -1,  0])
    y_ref = np.array([2, 0, 3, 1, 1, 4, 0])

    y = _dbscan_noisy2unique(x_ref)
    assert v_measure_score(y, y_ref) == 1

    # check inverse transform
    x = _dbscan_unique2noisy(y_ref)
    assert v_measure_score(x, x_ref) == 1


def test_binary_linkage2clusters():
    from freediscovery.cluster.utils import _binary_linkage2clusters
    from sklearn.metrics import v_measure_score
    n_samples = 10
    linkage = np.array([[1, 2],
                        [2, 3],
                        [5, 7],
                        [6, 9]])

    cluster_id = _binary_linkage2clusters(linkage, n_samples)

    cluster_id_ref = np.array([0, 1, 1, 1, 2, 3, 4, 3, 5, 4])

    assert cluster_id.shape == cluster_id_ref.shape
    # i.e. same clusters
    assert v_measure_score(cluster_id, cluster_id_ref) == 1.0


def test_merge_clusters():
    from freediscovery.cluster.utils import _merge_clusters

    X = np.array([[1, 2, 7, 9, 7, 8]]).T

    y = _merge_clusters(X)
    assert_equal(X, y[:, None])

    X_new = np.concatenate((X, X, X, X), axis=1)
    y = _merge_clusters(X_new)
    assert_equal(X, y[:, None])

    X = np.array([[1, 1, 2, 2, 3, 1, 3],
                  [2, 4, 2, 5, 1, 1, 3]]).T
    y = _merge_clusters(X)
    assert_equal(y, [1, 1, 1, 1, 3, 1, 3])


def test_select_top_words():
    try:
        import nltk
    except ImportError:
        raise SkipTest
    words_list = ['apple', 'apples', 'test', 'go']
    n_words = 2
    res = select_top_words(words_list, n=n_words)
    assert len(res) == n_words
    assert res == ['apple', 'test']


def test_birch_clusterig_single_nodes():

    basename = os.path.dirname(__file__)
    X = np.load(os.path.join(basename, '..', 'data',
                'ds_lsi_birch', 'data.npy'))
    branching_factor = 5

    mod = Birch(n_clusters=None, threshold=0.1,
                branching_factor=branching_factor, compute_labels=False,
                compute_sample_indices=True)
    mod.fit(X)

    htree, n_subclusters = birch_hierarchy_wrapper(mod)

    # let's compute cluster similarity
    for row in htree.flatten():
        inertia, S_sim = centroid_similarity(X,
                                             row['document_id_accumulated'])
        row['document_similarity'] = S_sim
        row['cluster_similarity'] = inertia

    assert htree.tree_size == n_subclusters

    doc_count = 0
    for el in htree.flatten():
        doc_count += len(el['document_id'])
        el.current_depth
        el.document_id_accumulated
    assert doc_count == len(htree['document_id_accumulated'])
    assert doc_count == X.shape[0]
    assert htree.document_count == X.shape[0]

    # make sure that we have no clusters with a single child
    assert sum(len(el.children) == 1 for el in htree.flatten()) == 0



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path
from numpy.testing import assert_allclose, assert_array_equal

from freediscovery.io import (parse_ground_truth_file,
                              parse_smart_tokens)


def test_parse_ground_truth_file():
    basename = os.path.dirname(__file__)
    filename = os.path.join(basename, "..","data", "ds_001", "ground_truth_file.txt")
    res = parse_ground_truth_file(filename)
    assert_allclose(res.is_relevant.values, [1, 1, 1, 0, 0, 0])


def test_parse_smart_stemmed():
    from textwrap import dedent
    text = dedent("""\
    .I 26187
    .W
    sunday complet tabulat race race race top top andret brazil brazil brazil rahal bobby lola lola lola cosworth cosworth cosworth cosworth ford ford ford ford mph reynard reynard reynard reynard
    reynard reynard merced merced merced christ benz benz benz motor motor fittipald jimmy vass al uns grand prix finish finish countr chass gil fer pensk bryan hert michael adrian gordon goodyear
    lap robby driv kph canad hond hond hond vancouv vancouv de indycar indycar scot fernandez jr

    .I 26188
    .W
    sunday won race race andret andret rahal bobby lola lola ford ford reynard merced christ benz motor fittipald grand prix finish finish michael win vancouv vancouv indycar indycar
    """)

    res = parse_smart_tokens(text)
    assert_array_equal(res.index, [26187, 26188])
    assert len(res.loc[26188].W.split(' ')) == 28
    res_ref = ['sunday', 'won', 'race', 'race', 'andret', 'andret', 'rahal',
               'bobby', 'lola', 'lola', 'ford', 'ford', 'reynard', 'merced',
               'christ', 'benz', 'motor', 'fittipald', 'grand', 'prix',
               'finish', 'finish', 'michael', 'win', 'vancouv', 'vancouv',
               'indycar', 'indycar']
    assert res.loc[26188].W.split(' ') == res_ref



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import (assert_allclose, assert_equal)

import pytest
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import normalize
from sklearn.externals import joblib

from freediscovery.neighbors import (NearestNeighborRanker,
                                     NearestCentroidRanker,
                                     _chunk_kneighbors)
from .run_suite import check_cache


basename = os.path.dirname(__file__)


cache_dir = check_cache()


@pytest.mark.parametrize('n_categories', [1, 2, 3, 4])
def test_nearest_neighbor_ranker(n_categories):
    # check that we have sensible results with respect to
    # NN1 binary classification (supervised, with both positive
    # and negative samples)
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.neighbors import NearestNeighbors
    np.random.seed(0)

    n_samples = 110
    n_features = 10

    X = np.random.rand(n_samples, n_features)
    normalize(X, copy=False)
    index = np.arange(n_samples, dtype='int')
    y = np.random.randint(0, n_categories, size=(n_samples,))
    index_train, index_test, y_train, y_test = train_test_split(index, y)
    X_train = X[index_train]
    X_test = X[index_test]

    rk = NearestNeighborRanker()
    rk.fit(X_train, y_train)
    y_pred, idx_pred = rk.kneighbors(X_test, batch_size=90) # choose a batch size smaller
                                                           # than n_samples

    assert y_pred.shape == (X_test.shape[0], n_categories)
    assert y_pred.min() >= -1 and y_pred.max() <= 1 # as we are using cosine similarities
    assert idx_pred.shape == (X_test.shape[0], n_categories)

    # postive scores correspond to positive documents

    #assert_equal((y_pred > 0), y_train[idx])

    cl = KNeighborsClassifier(n_neighbors=1, algorithm='brute', metric='cosine')
    cl.fit(X_train, y_train)

    idx_ref_cl = cl.predict(X_test)

    # make sure we get the same results as for the KNeighborsClassifier
    label_pred = np.argmax(y_pred, axis=1)
    assert_equal(label_pred, idx_ref_cl)

    nn = NearestNeighbors(n_neighbors=1, algorithm='brute', metric='cosine')
    nn.fit(X_train)
    S_ref_nn, idx_ref_nn = nn.kneighbors(X_test)

    assert_equal(idx_pred[range(len(label_pred)), label_pred], idx_ref_nn[:,0])
    assert_allclose(np.max(y_pred, axis=1)[:, None], 1 - S_ref_nn)


def test_nearest_neighbor_ranker_is_picklable():
    mod = NearestNeighborRanker()

    mod.fit([[0, 1], [1, 0]], [0, 1])

    try:
        tmp_file = os.path.join(cache_dir, 'tmp_NearestNeighborRanker.pkl')
        joblib.dump(mod, tmp_file)

        mod2 = joblib.load(tmp_file)
    finally:
        if os.path.exists(tmp_file):
            os.remove(tmp_file)


def test_nearest_centroid_ranker():
    # in the case where there is a single point by centroid,
    # nearest centroid should reduce to nearest neighbor
    from sklearn.neighbors import NearestNeighbors
    np.random.seed(0)

    n_samples = 100
    n_features = 120
    X = np.random.rand(n_samples, n_features)
    normalize(X, copy=False)
    index = np.arange(n_samples, dtype='int')
    y = np.arange(n_samples, dtype='int')
    index_train, index_test, y_train, y_test = train_test_split(index, y)
    X_train = X[index_train]
    X_test = X[index_test]


    nn = NearestNeighbors(n_neighbors=1, algorithm='brute')
    nn.fit(X_train)
    dist_ref, idx_ref = nn.kneighbors(X_test)

    nc = NearestCentroidRanker()
    nc.fit(X_train, y_train)
    dist_pred = nc.decision_function(X_test)
    y_pred = nc.predict(X_test)

    # ensures that we have the same number of unique ouput points
    # (even if absolute labels are not preserved)
    assert np.unique(idx_ref[:,0]).shape ==  np.unique(y_pred).shape

    assert_allclose(dist_pred, dist_ref[:,0])


@pytest.mark.parametrize('batch_size', [1, 101, 100])
def test_nn_chunking(batch_size):

    shape = (1000, 10)

    X = 2*np.ones(shape)
    def func(X):
        assert X.shape[0] > 0 # we need at least one point
        return X**2, X - 1

    d, idx = _chunk_kneighbors(func, X, batch_size=batch_size)

    assert_allclose(d, 4*np.ones(shape))
    assert_allclose(idx, 1*np.ones(shape))



# -*- coding: utf-8 -*-

"""jwzthreading.py

Contains an implementation of an algorithm for threading mail
messages, as described at http://www.jwz.org/doc/threading.html.

To use:

  Create a bunch of Message instances, one per message to be threaded,
  filling in the .subject, .message_id, and .references attributes.
  You can use the .message attribute to record the RFC-822 message object,
  or some other piece of information for your own purposes.

  Call the thread() function with a list of the Message instances.

  You'll get back a {subject line -> Container} dictionary; each
  container may have a .children attribute giving descendants of each
  message.  You'll probably want to sort these children by date, subject,
  or some other criterion.

Copyright (c) 2003-2016, A.M. Kuchling.

This code is under a BSD-style license; see the LICENSE file for details.
"""

from __future__ import print_function
from collections import deque, OrderedDict
import re
import sys

__all__ = ['Message', 'thread']

__version__ = "0.96"

#
# constants
#

MSGID_RE = re.compile(r'<([^>]+)>')
SUBJECT_RE = re.compile(
    r'((Re(\[\d+\])?:) | (\[ [^]]+ \])\s*)+', re.I | re.VERBOSE)


#
# models
#

class Container(dict):
    """Contains a tree of objects. Each container is a subclassed dict
    where the contents are stored.

    Attributes:
        children ([Container]): Possibly-empty list of child containers
        parent (Container): Parent container, if any
    """
    def __init__(self, **args):
        dict.__init__(self, **args)
        self.parent = None
        self.children = []

    def __repr__(self):
        return '<%s %x: %r>' % (self.__class__.__name__, id(self),
                                dict.__repr__(self))

    def __hash__(self):
        """ Make the container hashable. Care must be taken though not to change
        the container contents after the initialization as otherwise the hash
        value will change
        """
        return hash(tuple(sorted(self.items())) + (self.parent,))

    @property
    def is_dummy(self):
        """Check if the container has some content."""
        return not len(self.keys())

    def add_child(self, child):
        """Add a child to the container

        Parameters
        ----------
        child : Container
           Child to add.
        """
        if child.parent:
            child.parent.remove_child(child)
        self.children.append(child)
        child.parent = self

    def remove_child(self, child):
        """Remove a child from the container

        Parameters
        ----------
        child : Container
           Child to remove.
        """
        self.children.remove(child)
        child.parent = None

    def has_descendant(self, ctr):
        """Check if `ctr` is a descendant of this container.

        Parameters
        ----------
        ctr : Container
           possible descendant container.

        Returns
        -------
        True if `ctr` is a descendant of `self`, else False.
        """
        # To avoid recursing indefinitely, we'll do a depth-first
        # search; 'seen' tracks the containers we've already seen,
        # and 'stack' is a deque containing containers that we need
        # to look at.
        stack = deque()
        stack.append(self)
        seen = set()

        while stack:
            node = stack.pop()
            if node is ctr:
                return True
            seen.add(node)
            for child in node.children:
                if child not in seen:
                    stack.append(child)

        return False

    @property
    def tree_size(self):
        """Recursively count the number of children containers.
        The current container is also included in the count.
        """

        return 1 + sum([child.tree_size for child in self.children])

    @property
    def current_depth(self):
        """Compute the depth in the hierarchy of the current container"""

        if self.parent is None:
            return 0
        else:
            return 1 + self.parent.current_depth

    def flatten(self):
        """ Return a flatten version of the hierarchical tree

        Returns
        -------
        list : Containers
          a flat list of containers
        """
        from itertools import chain

        in_list = [[self]] + [child.flatten() for child in self.children]

        return list(chain.from_iterable(in_list))

    @property
    def root(self):
        """ Get the root container

        Returns
        -------
        Containe: the top most level container
        """

        if self.parent is None:
            return self
        else:
            return self.parent.root


class JwzContainer(Container):
    def collapse_empty(self, inplace=True):
        """ Collapse empty top level containers.

        This only applies to the JWZ algorithm

        If multiple messages reference a non existing top level message,
        by default JWZ threading algorithm will create a en empty top level
        container to be used as the root node.

        This method removes this empty container and makes the first child
        to be the root message. The other messages at depth == 1 then become
        it's children.

        Parameters
        ----------

        inplace : bool, default=True
           if True the original container is modified
        """

        if not inplace:
            raise NotImplementedError

        if 'message' not in self:
            raise ValueError("This method is only valid when used for "
                             "email threading")

        if self['message'] is not None:
            # nothing to be done
            return self

        if any([el['message'] is None for el in self.children]):
            raise ValueError('Children containers cannot be empty!')

        # In the following, self.message is None

        # make the 1st children the new root container
        children = self.children

        new_root = children[0]
        new_root.parent = None

        for idx in range(1, len(children)):
            child = children[idx]
            child.parent = new_root
            new_root.children.append(child)

        return new_root

    def to_dict(self, include=[]):
        """ Convert a Container tree to a nested dict
        """
        if 'message' not in self:
            raise ValueError('This method is currently valid with email'
                             'threading, please overwrite it for '
                             'other applications')

        if self['message'] is None:
            raise ValueError('Containers with None messages are not '
                             'supported:!\n this: {}'.format(self))

        res = {'id': self['message'].message_idx}

        for key in include:
            res[key] = getattr(self['message'], key)

        if self.parent is not None:
            if self.parent['message'] is not None:
                res['parent'] = self.parent['message'].message_idx
            else:
                raise ValueError('Containers with None messages are not '
                                 'supported:!\n    this: {}\n    parent: {}'
                                 .format(self, self.parent))
        else:
            res['parent'] = None

        res['children'] = [el.to_dict(include=include) for el in self.children]

        return res


class Message(object):
    """Represents a message to be threaded.

    Attributes:
        subject (str): Subject line of the message.
        message_id (str): Message ID as retrieved from the Message-ID
            header.
        references ([str]): List of message IDs from the In-Reply-To
            and References headers.
        message (any): Can contain information for the caller's use
            (e.g. an RFC-822 message object).
    """
    message = None
    message_id = None
    references = []
    subject = None

    message_idx = None  # internal message number in the mailbox

    def __init__(self, msg=None, message_idx=None, decode_header=False):
        if msg is None:
            return

        if message_idx is not None:
            self.message_idx = message_idx

        msg_id = MSGID_RE.search(msg.get('Message-ID', ''))
        if msg_id is None:
            raise ValueError('Message does not contain a Message-ID: header')

        self.message = msg
        self.message_id = msg_id.group(1)

        self.references = unique(MSGID_RE.findall(msg.get('References', '')))
        subject = msg.get('Subject', "No subject")
        if decode_header:
            from email.header import decode_header
            subject, subject_encoding = decode_header(subject)[0]
            if isinstance(subject, bytes) and subject_encoding is not None:
                if sys.version_info > (3, 0):
                    if subject_encoding == 'unknown-8bit':
                        try:
                            subject = subject.decode('utf-8')
                        except:
                            pass

                    else:
                        subject = subject.decode(subject_encoding)

        self.subject = subject

        # Get In-Reply-To: header and add it to references
        msg_id = MSGID_RE.search(msg.get('In-Reply-To', ''))
        if msg_id:
            msg_id = msg_id.group(1)
            if msg_id not in self.references:
                self.references.append(msg_id)

    def __repr__(self):
        return '<%s: %r>' % (self.__class__.__name__, self.message_id)

#
# functions
#

def unique(alist):
    result = OrderedDict()
    return [result.setdefault(e, e) for e in alist if e not in result]


def prune_container(container):
    """Prune a tree of containers.

    Recursively prune a tree of containers, as described in step 4 of
    the algorithm. Returns a list of the children that should replace
    this container.

    Arguments:
        container (Container): Container to prune

    Returns:
        List of zero or more containers.
    """
    # Prune children, assembling a new list of children
    new_children = []

    for ctr in container.children[:]:  # copy the container.children list
        pruned_child = prune_container(ctr)
        new_children.extend(pruned_child)
        container.remove_child(ctr)

    for child in new_children:
        container.add_child(child)

    if container.get('message') is None and not len(container.children):
        # step 4 (a) - nuke empty containers
        return []
    elif container.get('message') is None and (
        len(container.children) == 1 or container.parent is not None):
        # step 4 (b) - promote children
        children = container.children[:]
        for child in children:
            container.remove_child(child)
        return children
    else:
        # Leave this node in place
        return [container]

def sort_threads(threads, key='message_idx', missing=-1, reverse=False):
    """Sort threaded emails based on their root element

    Arguments:
        messages ([Container]): List of Container items
        group_by_subject (bool): Group root set by subject
               step 5 of the JWZ algorithm.
        key (str or None): optional sorting order for threads
               Valid values are "message_id", "subject", "message_idx" 
        missing (None): if the container has no message,
               replace it with this value
        reverse (book): reverse the order
    Returns:
        list ([Container]): sorted list of containers
    """

    def _sort_func(el):

        if el.get('message') is None:
            val = missing
        else:
            val = getattr(el.get('message'), key)
        if val is None:
            val = missing
        return val

    if key in ['message_id', 'subject', 'message_idx']:
        threads = sorted(threads, key=_sort_func)
    else:
        raise ValueError('Wrong input argument `sort_by`={}'.format(key))
    return threads


def thread(messages, group_by_subject=True):
    """Thread a list of mail items.

    Takes a list of Message objects, and returns a list of Containers.
    Containers are trees, with the `children` attribute containing
    a list of subtrees, so callers can then sort children by date
    or poster or whatever.

    Note: container ordering is not guaranteed by default,
    use the sort_threads function


    Arguments:
        messages ([Message]): List of Message items
        group_by_subject (bool): Group root set by subject
               (optional) step 5 of the JWZ algorithm.

    Returns:
        list of containers, sorted by date
    """
    # step one
    id_table = OrderedDict()

    for msg in messages:
        # step one (a)
        this_container = id_table.get(msg.message_id, None)
        if this_container is not None:
            this_container['message'] = msg
        else:
            this_container = JwzContainer(message=None)
            this_container['message'] = msg
            id_table[msg.message_id] = this_container

        # step one (b)
        prev = None
        for ref in msg.references:
            ## print "Processing reference for "+repr(msg.message_id)+": "+repr(ref)
            container = id_table.get(ref, None)
            if container is None:
                container = JwzContainer(message=None)
                id_table[ref] = container

            if prev is not None:
                #If they are already linked, don't change the existing links.
                if container.parent != None:
                    pass
                # Don't add link if it would create a loop
                elif container is this_container or \
                     container.has_descendant(prev) or \
                     prev.has_descendant(container):
                    pass
                else:
                    prev.add_child(container)

            prev = container
            ## print "Finished processing reference for "+repr(msg.message_id)+", container now: "
            ## print_container(container, 0, True)
        #1C
        if prev is not None:
            ##print "Setting parent of "+repr(this_container)+", to last reference: " + repr (prev)
            prev.add_child(this_container)
        else:
            if(this_container.parent):
                this_container.parent.remove_child(this_container)
        

    # step two - find root set
    root_set = [container for container in id_table.values()
                if container.parent is None]

    # step three - delete id_table
    del id_table

    # step four - prune empty containers
    for container in root_set:
        assert container.parent == None

    new_root_set = []
    for container in root_set:
        new_container = prune_container(container)
        new_root_set.extend(new_container)

    root_set = new_root_set

    # print '\n\nafter'
    # for ctr in root_set:
    # print_container(ctr)

    if not group_by_subject:
        # skip the following step
        return root_set

    # step five - group root set by subject
    subject_table = OrderedDict()
    for container in root_set:
        if container['message']:
            subj = container['message'].subject
        else:
            subj = container.children[0]['message'].subject

        subj = SUBJECT_RE.sub('', subj)
        if subj == '':
            continue

        existing = subject_table.get(subj, None)
        if (existing is None or
            (existing.message is not None and
             container.get('message') is None) or
            (existing.message is not None and
             container.get('message') is not None and
             len(existing.message.subject) > len(container['message'].subject))):
            subject_table[subj] = container


    # step five (c)
    for container in root_set:
        if container['message']:
            subj = container['message'].subject
        else:
            subj = container.children[0]['message'].subject

        subj = SUBJECT_RE.sub('', subj)
        ctr = subject_table.get(subj)

        if ctr is None or ctr is container:
            continue

        if ctr.is_dummy and container.is_dummy:
            for child in ctr.children:
                container.add_child(child)
        elif ctr.is_dummy or container.is_dummy:
            if ctr.is_dummy:
                ctr.add_child(container)
            else:
                container.add_child(ctr)
        elif len(ctr.message.subject) < len(container['message'].subject):
            # ctr has fewer levels of 're:' headers
            ctr.add_child(container)
        elif len(ctr.message.subject) > len(container['message'].subject):
            # container has fewer levels of 're:' headers
            container.add_child(ctr)
        else:
            new = JwzContainer(message=None)
            new.add_child(ctr)
            new.add_child(container)
            subject_table[subj] = new

    return list(subject_table.values())


def print_container(ctr, depth=0, debug=0):
    """Print summary of Thread to stdout."""
    if 'message' in ctr:
        if debug:
            message = repr(ctr) + ' ' + repr(ctr['message'] and ctr['message'].subject)
        else:
            message = str(ctr['message'] and ctr['message'].subject)
    else:
        message = str(ctr)

    print(''.join(['> ' * depth, message]))

    for child in ctr.children:
        print_container(child, depth + 1, debug)


def main():
    import mailbox
    import sys

    msglist = []

    print('Reading input file...')
    mbox = mailbox.mbox(sys.argv[1])
    for message in mbox:
        try:
            parsed_msg = Message(message)
        except ValueError:
            continue
        msglist.append(parsed_msg)

    print('Threading...')
    threads = thread(msglist)

    print('Output...')
    for container in threads:
        print_container(container)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os.path

from numpy.testing import assert_allclose, assert_equal

from sklearn.preprocessing import normalize
import pytest


from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.engine.ingestion import DocumentIndex
from freediscovery.exceptions import WrongParameter
from freediscovery.tests.run_suite import check_cache

basename = os.path.dirname(__file__)
data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")


def test_lsi():

    cache_dir = check_cache()
    n_components = 2

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mode='w')
    lsi_res, exp_var = lsi.fit_transform(n_components=n_components, alpha=1.0)
    assert lsi_res.components_.shape[0] == 5
    assert lsi_res.components_.shape[1] == fe.n_features_
    assert lsi._load_pars() is not None
    lsi._load_model()
    X_lsi = lsi._load_features()

    assert_allclose(normalize(X_lsi), X_lsi)

    lsi.list_models()
    lsi.delete()


def test_lsi_append_documents():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mode='w')
    lsi_res, exp_var = lsi.fit_transform(n_components=2, alpha=1.0)
    X_lsi = lsi._load_features()
    n_samples = fe.n_samples_

    docs = DocumentIndex.from_folder(data_dir).data
    docs['document_id'] += 10
    dataset_definition = docs[['file_path', 'document_id']].to_dict(orient='records')
    for row in dataset_definition:
        row['file_path'] = os.path.join(data_dir, row['file_path'])
    fe.append(dataset_definition)

    X_lsi_new = lsi._load_features()
    assert X_lsi_new.shape[0] == X_lsi.shape[0]*2
    assert_equal(X_lsi_new[:n_samples], X_lsi_new[:n_samples])


def test_lsi_remove_documents():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mode='w')
    lsi_res, exp_var = lsi.fit_transform(n_components=2, alpha=1.0)
    X_lsi = lsi._load_features()

    docs = DocumentIndex.from_folder(data_dir).data
    dataset_definition = docs[['document_id']].to_dict(orient='records')
    fe.remove([dataset_definition[2], dataset_definition[4]])

    X_lsi_new = lsi._load_features()
    assert X_lsi_new.shape[0] == X_lsi.shape[0] - 2


def test_custom_mid():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    mid_orig = "sklds"

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mid=mid_orig,
                      mode='w')
    lsi_res, exp_var = lsi.fit_transform(n_components=2, alpha=1.0)
    lsi._load_features()

    assert lsi.mid == mid_orig

    with pytest.raises(WrongParameter):
        lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mid=mid_orig,
                          mode='w')
        lsi.fit_transform(n_components=2, alpha=1.0)

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mid=mid_orig,
                      mode='fw')
    lsi.fit_transform(n_components=2, alpha=1.0)

    with pytest.raises(WrongParameter):
        lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mid='?',
                          mode='fw')
        lsi.fit_transform(n_components=2, alpha=1.0)



# -*- coding: utf-8 -*-

import os.path

import numpy as np
from numpy.testing import assert_allclose
import pytest

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.cluster import _ClusteringWrapper
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.tests.run_suite import check_cache


NCLUSTERS = 2


def fd_setup():
    basename = os.path.dirname(__file__)
    cache_dir = check_cache()
    np.random.seed(1)
    data_dir = os.path.join(basename, "..", "..",
                            "data", "ds_001", "raw")
    n_features = 110000
    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    dsid = fe.setup(n_features=n_features, use_hashing=False,
                    stop_words='english',
                    min_df=0.1, max_df=0.9)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=dsid, mode='w')
    lsi.fit_transform(n_components=6)
    return cache_dir, dsid, fe.filenames_, lsi.mid


def check_cluster_consistency(labels, terms):
    assert isinstance(labels, (list, np.ndarray))
    assert isinstance(terms, list)
    assert len(np.unique(labels)) == len(terms)


@pytest.mark.parametrize('method, use_lsi, args, cl_args',
                         [['k_means', None, {}, {}],
                          ['k_means', True,   {}, {}],
                          ['birch', True, {'threshold': 0.5}, {}],
                          ['birch', True, {'threshold': 0.5, 'branching_factor': 3, 'n_clusters': None}, {}],
                          #['ward_hc', True, {'n_neighbors': 5}, {}],
                          ['dbscan', False, {'eps': 0.5, 'min_samples': 2}, {}],
                          ['dbscan', True,   {'eps': 0.5, 'min_samples': 2}, {}]])
def test_clustering(method, use_lsi, args, cl_args):

    cache_dir, uuid, filenames, lsi_id = fd_setup()
    np.random.seed(1)
    n_top_words = 9

    if use_lsi:
        parent_id = lsi_id
    else:
        parent_id = uuid

    cat = _ClusteringWrapper(cache_dir=cache_dir, parent_id=parent_id)
    cm = getattr(cat, method)
    if 'n_clusters' not in args:
        args['n_clusters'] = NCLUSTERS
    labels = cm(**args)

    htree = cat._load_htree()

    mid = cat.mid

    if method == 'birch' and cat._pars['is_hierarchical']:
        assert htree != {}
        flat_tree = htree.flatten()

        terms = cat.compute_labels(n_top_words=n_top_words,
                                   cluster_indices=[row['document_id_accumulated']
                                                    for row in flat_tree])
        for label, row in zip(terms, flat_tree):
            row['cluster_label'] = label
    else:
        terms = cat.compute_labels(n_top_words=n_top_words, **cl_args)

        if method == 'ward_hc':
            assert sorted(htree.keys()) == sorted(['n_leaves',
                                                   'n_components', 'children'])
        else:
            assert not htree

        if method == 'dbscan':
            assert (labels != -1).all()

        check_cluster_consistency(labels, terms)
        cat.scores(np.random.randint(0, NCLUSTERS-1, size=len(labels)), labels)
        # load the model saved to disk
        km = cat._load_model()
        assert_allclose(labels, km.labels_)
        if method != 'dbscan':
            # DBSCAN does not take the number of clusters as input
            assert len(terms) == NCLUSTERS
            assert len(np.unique(labels)) == NCLUSTERS

        for el in terms:
            assert len(el) == n_top_words
        cluster_indices = np.nonzero(labels == 0)
        if use_lsi:
            # use_lsi=False is not supported for now
            terms2 = cat.compute_labels(cluster_indices=[cluster_indices],
                                        **cl_args)
            # 70% of the terms at least should match
            if method != 'dbscan':
                assert sum([el in terms[0]
                            for el in terms2[0]]) > 0.7*len(terms2[0])

    # make sure we can load it
    cat2 = _ClusteringWrapper(cache_dir=cache_dir, mid=mid)
    cat.delete()



"""
Clustering
==========

Cluster documents into clusters
"""

import os.path
import pandas as pd
from time import time
import requests

pd.options.display.float_format = '{:,.3f}'.format


dataset_name = "treclegal09_2k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

###############################################################################
#
# 0. Load the example dataset
# ---------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()

# To use a custom dataset, simply specify the following variables
data_dir = input_ds['metadata']['data_dir']
dataset_definition = [{'document_id': row['document_id'],
                       'file_path': os.path.join(data_dir, row['file_path'])}
                      for row in input_ds['dataset']]

###############################################################################
#
# 1. Feature extraction (non hashed)
# ----------------------------------
# 1.a Load dataset and initalize feature extraction

url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))


###############################################################################
#
# 1.b Run feature extraction

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
res = requests.post(url, json={'dataset_definition': dataset_definition})


###############################################################################
#
# 2. Calculate LSI
# ----------------

url = BASE_URL + '/lsi/'
print("POST", url)

n_components = 300
res = requests.post(url,
                    json={'n_components': n_components,
                          'parent_id': dsid
                          }).json()

lsi_id = res['id']
print('  => LSI model id = {}'.format(lsi_id))
print(('  => SVD decomposition with {} dimensions '
       'explaining {:.2f} % variabilty of the data')
      .format(n_components, res['explained_variance']*100))

###############################################################################
#
# 3. Document Clustering (LSI + K-Means)
# --------------------------------------

print("\n3.a. Document clustering (LSI + K-means)")

url = BASE_URL + '/clustering/k-mean/'
print(" POST", url)
t0 = time()
res = requests.post(url,
                    json={'parent_id': lsi_id,
                          'n_clusters': 10,
                          }).json()

mid = res['id']
print("     => model id = {}".format(mid))

###############################################################################
#
# 3.b. Computing cluster labels

url = BASE_URL + '/clustering/k-mean/{}'.format(mid)
print(" GET", url)
res = requests.get(url,
                   json={'n_top_words': 3
                         }).json()
t1 = time()


data = res['data']
for row in data:
    row['n_documents'] = len(row.pop('documents'))

print('    .. computed in {:.1f}s'.format(t1 - t0))
print(pd.DataFrame(data))


###############################################################################
#
# 4. Document Clustering (LSI + Birch Clustering)
# -----------------------------------------------

print("\n4.a. Document clustering (LSI + Birch clustering)")

url = BASE_URL + '/clustering/birch/'
print(" POST", url)
t0 = time()
res = requests.post(url,
                    json={'parent_id': lsi_id,
                          'n_clusters': -1,
                          'min_similarity': 0.7,
                          'branching_factor': 20,
                          'max_tree_depth': 2,
                          }).json()

mid = res['id']
print("     => model id = {}".format(mid))

###############################################################################
#
# 4.b. Computing cluster labels

url = BASE_URL + '/clustering/birch/{}'.format(mid)
print(" GET", url)
res = requests.get(url,
                   json={'n_top_words': 3
                         }).json()
t1 = time()

print('    .. computed in {:.1f}s'.format(t1 - t0))
data = res['data']
for row in data:
    row['n_documents'] = len(row.pop('documents'))

print(pd.DataFrame(data))

###############################################################################
#
# 5. Optimal sampling (LSI + Birch Clustering)
# --------------------------------------------
t0 = time()
url = BASE_URL + '/clustering/birch/{}'.format(mid)
print(" GET", url)
res = requests.get(url,
                   json={'return_optimal_sampling': True,
                         'sampling_min_coverage': 0.9
                         }).json()
t1 = time()
print('    .. computed in {:.1f}s'.format(t1 - t0))
data = res['data']

print(pd.DataFrame(data))

###############################################################################
#
# 5. Delete the extracted features

url = BASE_URL + '/feature-extraction/{}'.format(dsid)
requests.delete(url)



# -*- coding: utf-8 -*-

from numpy.testing import assert_equal, assert_almost_equal

from .base import (parse_res, V01, app, app_notest, get_features, get_features_lsi,
                   email_data_dir)


#=============================================================================#
#
#                     Email Threading
#
#=============================================================================#


def test_api_thread_emails(app):

    dsid, pars, _ = get_features(app, parse_email_headers=True,
                                 dataset='fedora_ml_3k_subset')

    url = V01 + "/email-threading"
    pars = {'parent_id': dsid}

    data = app.post_check(url, json=pars)
    assert sorted(data.keys()) == sorted(['data', 'id'])
    mid = data['id']

    tree_ref = [{'id': 0, 'parent': None, 'children': [
                 {'id': 1, 'children': [], 'parent': 0},
                 {'id': 2, 'parent': 0,  'children': [
                        {'id': 3, 'children': [], 'parent': 2},
                        {'id': 4, 'children': [], 'parent': 2}]}]
                 }]

    def remove_subject_field(d):
        del d['subject']
        for el in d['children']:
            remove_subject_field(el)

    tree_res = data['data']
    for el in tree_res:
        remove_subject_field(el)

    assert data['data'] == tree_ref

    url += '/{}'.format(mid)
    data = app.get_check(url)
    assert sorted(data.keys()) == sorted(['group_by_subject'])

    app.delete_check(url)



"""Utilities for file download and caching, adapted from Keras 2

MIT Licence.
"""

import tarfile
import zipfile
import os
import sys
import time
import shutil
import hashlib

try:
    from urllib.error import URLError
    from urllib.error import HTTPError
except ImportError:
    from urllib2 import URLError, HTTPError
if sys.version_info[0] == 2:
    def urlretrieve(url, filename, reporthook=None, data=None):
        """Replacement for `urlretrive` for Python 2.
        Under Python 2, `urlretrieve` relies on `FancyURLopener` from legacy
        `urllib` module, known to have issues with proxy management.
        # Arguments
            url: url to retrieve.
            filename: where to store the retrieved data locally.
            reporthook: a hook function that will be called once
                on establishment of the network connection and once
                after each block read thereafter.
                The hook will be passed three arguments;
                a count of blocks transferred so far,
                a block size in bytes, and the total size of the file.
            data: `data` argument passed to `urlopen`.
        """
        from urllib2 import urlopen

        def chunk_read(response, chunk_size=8192, reporthook=None):
            content_type = response.info().get('Content-Length')
            total_size = -1
            if content_type is not None:
                total_size = int(content_type.strip())
            count = 0
            while 1:
                chunk = response.read(chunk_size)
                count += 1
                if not chunk:
                    reporthook(count, total_size, total_size)
                    break
                if reporthook:
                    reporthook(count, chunk_size, total_size)
                yield chunk

        response = urlopen(url, data)
        with open(filename, 'wb') as fd:
            for chunk in chunk_read(response, reporthook=reporthook):
                fd.write(chunk)
else:
    from urllib.request import urlretrieve

import numpy as np

INTERNAL_DATA_DIR = os.path.dirname(os.path.dirname((__file__)))
INTERNAL_DATA_DIR = os.path.abspath(os.path.join(INTERNAL_DATA_DIR, 'data'))


def _extract_archive(file_path, path='.', archive_format='auto'):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format is 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, str):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type is 'tar':
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type is 'zip':
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError,
                        KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def _get_file(fname,
              origin,
              md5_hash=None,
              file_hash=None,
              hash_algorithm='auto',
              extract=False,
              archive_format='auto',
              cache_dir=None):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `cache_dir/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        md5_hash: Deprecated in favor of 'file_hash'.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, tar.bz and tar.xz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the current dir.

    # Returns
        path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = os.path.abspath('.')
    elif isinstance(cache_dir, str):
        cache_dir = os.path.abspath(cache_dir)

    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = 'md5'

    if not os.path.exists(cache_dir):
        raise ValueError('cache_dir {} does not exist!'.format(cache_dir))


    if extract:
        extract_fpath = fname
        fpath = extract_fpath + '.tar.gz'
    else:
        extract_fpath = None
        fpath = fname

    download = False
    for base_dir in [cache_dir, INTERNAL_DATA_DIR]:
        fpath = os.path.join(base_dir, fpath)
        if os.path.exists(fpath):
            # File found; verify integrity if a hash was provided.
            if file_hash is not None:
                if not _validate_file(str(fpath), file_hash, algorithm=hash_algorithm):
                    print('A local file was found, but it seems to be '
                          'incomplete or outdated because the ' +
                          hash_algorithm +
                          ' file hash does not match the original value of ' +
                          file_hash + ' so we will re-download the data.')
                    download = True
            break
    else:
        base_dir = cache_dir
        download = True

    if download:
        print('Downloading data from', origin)

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size is -1:
                    total_size = None
                ProgressTracker.progbar = _Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = 'URL fetch failure on {}: {} -- {}'
        try:
            try:
                urlretrieve(origin, str(fpath), dl_progress)
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
        except (Exception, KeyboardInterrupt) as e:
            if fpath.exists():
                fpath.unlink()
            raise
        ProgressTracker.progbar = None

    if extract_fpath is not None:
        extract_fpath = os.path.join(base_dir, extract_fpath)

    print(' ')

    if extract:
        _extract_archive(str(fpath), str(base_dir), archive_format='tar')
        return extract_fpath

    return fpath


def _hash_file(fpath, algorithm='sha256', chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    # Example

    ```python
       >>> from keras.data_utils import _hash_file
       >>> _hash_file('/path/to/file.zip')
       'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        The file hash
    """
    if (algorithm is 'sha256') or (algorithm is 'auto' and len(hash) is 64):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


def _validate_file(fpath, file_hash, algorithm='auto', chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        Whether the file is valid
    """
    if ((algorithm is 'sha256') or
            (algorithm is 'auto' and len(file_hash) is 64)):
        hasher = 'sha256'
    else:
        hasher = 'md5'

    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False


class _Progbar(object):
    """Displays a progress bar.
    # Arguments
        target: Total number of steps expected, None if unknown.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05):
        self.width = width
        if target is None:
            target = -1
        self.target = target
        self.sum_values = {}
        self.unique_values = []
        self.start = time.time()
        self.last_update = 0
        self.interval = interval
        self.total_width = 0
        self.seen_so_far = 0
        self.verbose = verbose

    def update(self, current, values=None, force=False):
        """Updates the progress bar.
        # Arguments
            current: Index of current step.
            values: List of tuples (name, value_for_last_step).
                The progress bar will display averages for these values.
            force: Whether to force visual progress update.
        """
        values = values or []
        for k, v in values:
            if k not in self.sum_values:
                self.sum_values[k] = [v * (current - self.seen_so_far),
                                      current - self.seen_so_far]
                self.unique_values.append(k)
            else:
                self.sum_values[k][0] += v * (current - self.seen_so_far)
                self.sum_values[k][1] += (current - self.seen_so_far)
        self.seen_so_far = current

        now = time.time()
        if self.verbose == 1:
            if not force and (now - self.last_update) < self.interval:
                return

            prev_total_width = self.total_width
            sys.stdout.write('\b' * prev_total_width)
            sys.stdout.write('\r')

            if self.target is not -1:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%%%dd [' % (numdigits, numdigits)
                bar = barstr % (current, self.target)
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
                sys.stdout.write(bar)
                self.total_width = len(bar)

            if current:
                time_per_unit = (now - self.start) / current
            else:
                time_per_unit = 0
            eta = time_per_unit * (self.target - current)
            info = ''
            if current < self.target and self.target is not -1:
                info += ' - ETA: %ds' % eta
            else:
                info += ' - %ds' % (now - self.start)
            for k in self.unique_values:
                info += ' - %s:' % k
                if isinstance(self.sum_values[k], list):
                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self.sum_values[k]

            self.total_width += len(info)
            if prev_total_width > self.total_width:
                info += ((prev_total_width - self.total_width) * ' ')

            sys.stdout.write(info)
            sys.stdout.flush()

            if current >= self.target:
                sys.stdout.write('\n')

        if self.verbose == 2:
            if current >= self.target:
                info = '%ds' % (now - self.start)
                for k in self.unique_values:
                    info += ' - %s:' % k
                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                sys.stdout.write(info + "\n")

        self.last_update = now

    def add(self, n, values=None):
        self.update(self.seen_so_far + n, values)



# -*- coding: utf-8 -*-

from ...utils import dict2type

from .base import parse_res, V01, app, app_notest, get_features, get_features_lsi


def test_example_dataset(app):
    data = app.get_check(V01 + '/example-dataset/20_newsgroups_micro')
    assert dict2type(data, max_depth=1) == {'dataset': 'list',
                                            'training_set': 'list',
                                            'metadata': 'dict'}
    assert dict2type(data['metadata']) == {'data_dir': 'str',
                                           'name': 'str'}
    assert dict2type(data['training_set'][0]) == {'category': 'str',
                                               'document_id': 'int',
                                               'internal_id': 'int',
                                               'file_path': 'str'}
    assert dict2type(data['dataset'][0]) == {'category': 'str',
                                           'document_id': 'int',
                                           'internal_id': 'int',
                                           'file_path': 'str'
                                          }



def test_openapi_specs(app):
    data = app.get_check('/openapi-specs.json')
    assert data['swagger'] == '2.0'


def test_version(app):
    data = app.get_check('/')
    assert dict2type(data) == {'version': {'number': 'str'},
                               'env': {'python_version': 'str'},
                               'config': 'NoneType'}



"""
Duplicate Detection
===================

Find near-duplicates in a text collection
"""
from __future__ import print_function

from time import time

import pandas as pd
import requests

pd.options.display.float_format = '{:,.3f}'.format


dataset_name = "treclegal09_2k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

###############################################################################
#
# 0. Load the test dataset
# ------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()


# To use a custom dataset, simply specify the following variables
data_dir = input_ds['metadata']['data_dir']

###############################################################################
#
# 1. Feature extraction (non hashed)
# ----------------------------------

print("\n1.a Load dataset and initalize feature extraction")
url = BASE_URL + '/feature-extraction'
print(" POST", url)

fe_opts = {'weighting': 'ntc',
           'n_features': 30001,
           'min_df': 4, 'max_df': 0.75
           }
res = requests.post(url, json=fe_opts)

dsid = res.json()['id']
print("   => received {}".format(list(res.json().keys())))
print("   => dsid = {}".format(dsid))


###############################################################################
#
# 1.b Run feature extraction

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
res = requests.post(url, json={"data_dir": data_dir})

print("\n1.d. check the parameters of the extracted features")
url = BASE_URL + '/feature-extraction/{}'.format(dsid)
print(' GET', url)
res = requests.get(url)

data = res.json()
print('\n'.join(['     - {}: {}'.format(key, val)
      for key, val in data.items() if "filenames" not in key]))


###############################################################################
#
# 2. Compute LSI
# --------------

url = BASE_URL + '/lsi/'
print("POST", url)

n_components = 100
res = requests.post(url,
                    json={'n_components': n_components,
                          'parent_id': dsid
                          }).json()

lsi_id = res['id']

###############################################################################
#
# 3. Near Duplicates detection by cosine similarity (DBSCAN)
# ----------------------------------------------------------

url = BASE_URL + '/clustering/dbscan/'
print(" POST", url)
t0 = time()
res = requests.post(url,
                    json={'parent_id': lsi_id,
                          'min_similarity': 0.90,
                          'n_max_samples': 2
                          }).json()

mid = res['id']
print("     => model id = {}".format(mid))

url = BASE_URL + '/clustering/dbscan/{}'.format(mid)
print(" GET", url)
# don't compute cluster labels
res = requests.get(url, json={'n_top_words': 0}).json()
t1 = time()

print('    .. computed in {:.1f}s'.format(t1 - t0))

data = res['data']
print('Found {} duplicates / {}'
      .format(sum([len(row['documents'])
                   for row in data if len(row['documents']) > 1]),
              len(input_ds['dataset'])))


###############################################################################
#
# 4. Near Duplicates Detection using I-Match
# ------------------------------------------

url = BASE_URL + '/duplicate-detection/'
print(" POST", url)
t0 = time()
res = requests.post(url, json={'parent_id': dsid,
                               'method': 'i-match'})

data = res.json()
mid = data['id']
print("     => model id = {}".format(mid))

print('    .. computed in {:.1f}s'.format(time() - t0))


url = BASE_URL + '/duplicate-detection/{}'.format(mid)
print(" GET", url)
t0 = time()
res = requests.get(url, json={'n_rand_lexicons': 10,
                              'rand_lexicon_ratio': 0.9}).json()
t1 = time()
print('    .. computed in {:.1f}s'.format(time() - t0))

data = res['data']

print('Found {} duplicates / {}'
      .format(sum([len(row['documents']) for row in data]),
              len(input_ds['dataset'])))


###############################################################################
#
# 3. Duplicate detection by Simhash
# ---------------------------------

try:
    import simhash
    url = BASE_URL + '/duplicate-detection/'
    print(" POST", url)
    t0 = time()
    res = requests.post(url, json={'parent_id': dsid,
                                   'method': 'simhash'})

    data = res.json()
    mid = data['id']
    print("     => model id = {}".format(mid))

    print('    .. computed in {:.1f}s'.format(time() - t0))

    url = BASE_URL + '/duplicate-detection/{}'.format(mid)
    print(" GET", url)
    t0 = time()
    res = requests.get(url, json={'distance': 1})
    data = res.json()
    print('    .. computed in {:.1f}s'.format(time() - t0))

    data = data['data']

    print('Found {} duplicates / {}'
          .format(sum([len(row['documents']) for row in data]),
                  len(input_ds['dataset'])))
except ImportError:
    print("simhash is not installed or not supported "
          " (e.g. on Windows)")

###############################################################################
#
# 4 Delete the extracted features

url = BASE_URL + '/feature-extraction/{}'.format(dsid)
requests.delete(url)



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import sklearn


def _parse_version(version_string):
    """ adapted from sklearn """
    version = []
    for x in version_string.split('.'):
        try:
            version.append(int(x))
        except ValueError:
            # x may be of the form dev-1ea1592
            version.append(x)
    return tuple(version)


sklearn_version = _parse_version(sklearn.__version__)



# -*- coding: utf-8 -*-

import os.path
from numpy.testing import assert_equal, assert_array_equal
from pandas.util.testing import assert_frame_equal
import pytest
import pandas as pd

from freediscovery.engine.ingestion import DocumentIndex, _infer_document_id_from_path
from freediscovery.tests.run_suite import check_cache
from freediscovery.exceptions import (NotFound)

basename = os.path.dirname(__file__)
data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")
cache_dir = check_cache()


fnames_in = ['0.7.47.101442.txt',
             '0.7.47.117435.txt',
             '0.7.6.28635.txt',
             '0.7.6.28636.txt',
             '0.7.6.28637.txt',
             '0.7.6.28638.txt']
fnames_in_abs = [os.path.join(data_dir, el) for el in fnames_in]


def test_ingestion_base_dir():
    dbi = DocumentIndex.from_folder(data_dir)
    dbi._make_relative_paths()
    data_dir_res, filenames, db = dbi.data_dir, dbi.filenames_, dbi.data
    assert data_dir_res == os.path.normpath(data_dir)
    assert_array_equal(db.columns.values, ['file_path', 'internal_id', 'document_id'])
    assert_array_equal(db.file_path.values, fnames_in)
    assert_array_equal([os.path.normpath(os.path.join(data_dir_res, el))
                        for el in filenames],
                       [os.path.join(data_dir_res, el)
                        for el in db.file_path.values])


def test_search_2fields():
    dbi = DocumentIndex.from_folder(data_dir)
    dbi._make_relative_paths()

    query = pd.DataFrame([{'internal_id': 3},
                          {'internal_id': 1},
                          {'internal_id': 2}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [3, 1, 2])
    assert_array_equal(sorted(sres.columns),
                       sorted(['internal_id', 'file_path', 'document_id']))

    # make sure that if we have some additional field,
    # we still use the internal_id
    query = pd.DataFrame([{'internal_id': 1, 'a': 2},
                          {'internal_id': 2, 'b': 4},
                          {'internal_id': 1, 'a': 3}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [1, 2, 1])
    assert_array_equal(sorted(sres.columns),
                       sorted(['internal_id', 'file_path', 'document_id']))

    sres = dbi.search(query, drop=False)
    assert_equal(sres.internal_id.values, [1, 2, 1])
    assert_array_equal(sorted(sres.columns),
                       sorted(['internal_id', 'file_path', 'document_id',
                               'a', 'b']))

    query = pd.DataFrame([{'file_path': "0.7.6.28637.txt"},
                          {'file_path': "0.7.47.117435.txt"}])
    del dbi.data['file_path']
    sres = dbi.search(query)
    query_res = [dbi.data.file_path.values.tolist().index(el)
                 for el in query.file_path.values]
    assert_array_equal(query_res, sres.internal_id)


def test_search_not_found():
    dbi = DocumentIndex.from_folder(data_dir)
    query = pd.DataFrame([{'file_path': "DOES_NOT_EXISTS"},
                          {'file_path': "0.7.6.28637.txt"}])
    with pytest.raises(NotFound):
        sres = dbi.search(query)


@pytest.mark.parametrize('return_file_path',
                         ['return_file_path', 'dont_return_file_path'])
def test_ingestion_render(return_file_path):

    def _process_results(rd):
        rd = pd.DataFrame(rd)
        if return_file_path:
            assert 'file_path' in rd.columns
            del rd['file_path']
        return rd

    # make it a binary variable
    return_file_path = (return_file_path == 'return_file_path')

    md = [{'file_path': '/test',  'document_id': 2},
          {'file_path': '/test2', 'document_id': 1},
          {'file_path': '/test3', 'document_id': 7},
          {'file_path': '/test8', 'document_id': 9},
          {'file_path': '/test9', 'document_id': 4}]

    for idx, el in enumerate(md):
        el['internal_id'] = idx

    dbi = DocumentIndex.from_list(md)
    query = pd.DataFrame([{'a': 2, 'internal_id': 3},
                          {'a': 4, 'internal_id': 1}])
    res = pd.DataFrame([{'a': 2, 'internal_id': 3, 'document_id': 9},
                        {'a': 4, 'internal_id': 1, 'document_id': 1}])

    rd = dbi.render_dict(query, return_file_path=return_file_path)
    rd = _process_results(rd)
    assert_frame_equal(rd, res)
    rd = dbi.render_dict(return_file_path=return_file_path)
    rd = _process_results(rd)
    assert_frame_equal(rd.loc[[0]],
                       pd.DataFrame([{'internal_id': 0, 'document_id': 2}]))
    assert len(rd) == len(md)

    rd = dbi.render_list(res, return_file_path=return_file_path)
    rd = _process_results(rd)
    assert sorted(rd.keys()) == sorted(['internal_id', 'document_id', 'a'])
    assert_frame_equal(pd.DataFrame(rd),
                       pd.DataFrame([{'a': 2, 'internal_id': 3, 'document_id': 9},
                                     {'a': 4, 'internal_id': 1, 'document_id': 1}]),
                       check_like=True)

    rd = dbi.render_list()
    assert sorted(rd.keys()) == sorted(['internal_id', 'document_id'])


def test_search_document_id():
    md = [{'file_path': '/test',  'document_id': 2},
          {'file_path': '/test2', 'document_id': 1},
          {'file_path': '/test3', 'document_id': 7},
          {'file_path': '/test8', 'document_id': 9},
          {'file_path': '/test9', 'document_id': 4}]

    for idx, el in enumerate(md):
        el['internal_id'] = idx

    dbi = DocumentIndex.from_list(md)
    dbi._make_relative_paths()
    query = pd.DataFrame([{'internal_id': 1},
                          {'internal_id': 2},
                          {'internal_id': 1}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [1, 2, 1])
    assert_array_equal(sorted(sres.columns),
                       sorted(['internal_id', 'file_path', 'document_id']))

    # make sure we use internal id first
    query = pd.DataFrame([{'internal_id': 1, 'document_id': 2},
                          {'internal_id': 2, 'document_id': 2},
                          {'internal_id': 1, 'document_id': 2}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [1, 2, 1])

    query = pd.DataFrame([{'document_id': 4},
                          {'document_id': 9},
                          {'document_id': 2}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [4, 3, 0])


def test_search_document_rendition_id():
    md = [{'file_path': '/test',  'document_id': 0, 'rendition_id': 0},
          {'file_path': '/test2', 'document_id': 0, 'rendition_id': 1},
          {'file_path': '/test3', 'document_id': 1, 'rendition_id': 0},
          {'file_path': '/test8', 'document_id': 2, 'rendition_id': 0},
          {'file_path': '/test9', 'document_id': 3, 'rendition_id': 0}]

    for idx, el in enumerate(md):
        el['internal_id'] = idx

    # can always index with internal_id
    dbi = DocumentIndex.from_list(md)
    dbi._make_relative_paths()

    query = pd.DataFrame([{'internal_id': 1},
                          {'internal_id': 2},
                          {'internal_id': 1}])
    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [1, 2, 1])
    assert_array_equal(sorted(sres.columns),
                       sorted(['internal_id', 'file_path',
                               'document_id', 'rendition_id']))

    # the internal id is not sufficient to fully index documents in this case
    query = pd.DataFrame([{'document_id': 0},
                          {'document_id': 1},
                          {'document_id': 2}])
    with pytest.raises(ValueError):
        sres = dbi.search(query)

    query = pd.DataFrame([{'document_id': 0, 'rendition_id': 0},
                          {'document_id': 1, 'rendition_id': 0},
                          {'document_id': 2, 'rendition_id': 0}])

    sres = dbi.search(query)
    assert_equal(sres.internal_id.values, [0, 2, 3])


def test_bad_search_document_rendition_id():
    md = [{'file_path': '/test',  'document_id': 0, 'rendition_id': 0},
          {'file_path': '/test2', 'document_id': 0, 'rendition_id': 1},
          {'file_path': '/test3', 'document_id': 1, 'rendition_id': 0},
          {'file_path': '/test8', 'document_id': 2, 'rendition_id': 0},
          {'file_path': '/test9', 'document_id': 3, 'rendition_id': 0}]
    for idx, el in enumerate(md):
        el['internal_id'] = idx

    # can always index with internal_id
    dbi = DocumentIndex.from_list(md)
    query = pd.DataFrame([{'internal_id': 1},
                          {'internal_id': 2},
                          {'document_id': 1}])
    with pytest.raises(NotFound):
        sres = dbi.search(query)


def test_ingestion_pickling():
    from sklearn.externals import joblib
    db = DocumentIndex.from_folder(data_dir)
    fname = os.path.join(cache_dir, 'document_index')
    # check that db is picklable
    joblib.dump(db, fname)
    db2 = joblib.load(fname)
    os.remove(fname)


@pytest.mark.parametrize('n_fields', [1, 2, 3])
def test_ingestion_metadata(n_fields):
    metadata = []
    for idx, fname in enumerate(fnames_in_abs):
        el = {'file_path': fname}
        if n_fields >= 2:
            el['document_id'] = 'a' + str(idx + 100)
        if n_fields >= 3:
            el['rendition_id'] = 1
        metadata.append(el)

    dbi = DocumentIndex.from_list(metadata)
    dbi._make_relative_paths()
    data_dir_res, filenames, db = dbi.data_dir, dbi.filenames_, dbi.data

    if n_fields in [1, 2]:
        columns_ref = sorted(['file_path', 'document_id', 'internal_id'])
    elif n_fields == 3:
        columns_ref = sorted(['file_path', 'document_id', 'rendition_id',
                              'internal_id'])

    assert_array_equal(sorted(db.columns.values), columns_ref)
    assert_array_equal([os.path.normpath(os.path.join(data_dir_res, el))
                        for el in filenames],
                       [os.path.join(data_dir_res, el)
                        for el in db.file_path.values])


def test_infer_document_id_from_path():
    file_path = ['test/file1.txt', 'test3/file2.txt', 'test4/file4.txt']

    document_id = _infer_document_id_from_path(file_path)
    assert_array_equal(document_id, [1, 2, 4])

    file_path = ['test/file.txt', 'test3/file2.txt', 'test4/file4.txt']

    document_id = _infer_document_id_from_path(file_path)
    assert document_id is None

    file_path = ['test/file1.txt', 'test3/file1.txt', 'test4/file4.txt']

    document_id = _infer_document_id_from_path(file_path)
    assert document_id is None
